"""
This module defines an ASE interface to Turbomole: http://www.turbomole.com/

QMMM functionality provided by Markus Kaukonen <markus.kaukonen@iki.fi>.

Please read the license file (../../LICENSE)

Contact: Ivan Kondov <ivan.kondov@kit.edu>
"""
import os
import re
import warnings
from subprocess import Popen, PIPE
from math import log10, floor
import numpy as np
from ase import Atoms
from ase.units import Ha, Bohr
from ase.io import read, write
from ase.calculators.calculator import FileIOCalculator
from ase.calculators.calculator import PropertyNotImplementedError, ReadError
from ase.utils import basestring


def read_output(regex):
    """collects all matching strings from the output"""
    hitlist = []
    checkfiles = []
    for filename in os.listdir('.'):
        if filename.startswith('job.') or filename.endswith('.out'):
            checkfiles.append(filename)
    for filename in checkfiles:
        with open(filename, 'rt') as f:
            lines = f.readlines()
            for line in lines:
                match = re.search(regex, line)
                if match:
                    hitlist.append(match.group(1))
    return hitlist


def execute(args, input_str=None, error_test=True,
            stdout_tofile=True):
    """executes a turbomole executable and process the outputs"""

    if isinstance(args, basestring):
        args = args.split()

    if stdout_tofile:
        stdout_file = 'ASE.TM.' + args[0] + '.out'
        stdout = open(stdout_file, 'w')
    else:
        stdout = PIPE

    if input_str:
        stdin = input_str.encode()
    else:
        stdin = None

    message = 'TM command "' + args[0] + '" execution failed'
    try:
        proc = Popen(args, stdin=PIPE, stderr=PIPE, stdout=stdout)
        res = proc.communicate(input=stdin)
        if error_test:
            error = res[1].decode()
            if 'abnormally' in error or 'ended normally' not in error:
                message += ' with error:\n' + error
                message += '\nSee file ' + stdout_file + ' for details.\n'
                raise RuntimeError(message)
    except RuntimeError as err:
        raise err
    except OSError as err:
        raise OSError(err.args[1] + '\n' + message)
    else:
        print('TM command: "' + args[0] + '" successfully executed')

    if not stdout_tofile:
        return res[0].decode()


def add_data_group(data_group, string=None, raw=False):
    """write a turbomole data group to control file"""
    if raw:
        data = data_group
    else:
        data = '$' + data_group
        if string:
            data += ' ' + string
        data += '\n'
    f = open('control', 'r+')
    lines = f.readlines()
    f.seek(0)
    f.truncate()
    lines.insert(2, data)
    f.write(''.join(lines))
    f.close()


def read_data_group(data_group):
    """read a turbomole data group from control file"""
    args = ['sdg', data_group]
    dg = execute(args, error_test=False, stdout_tofile=False)
    return dg.strip()


def delete_data_group(data_group):
    """delete a turbomole data group from control file"""
    command = ['kdg', data_group]
    execute(command, error_test=False, stdout_tofile=False)


class TurbomoleOptimizer:
    def __init__(self, atoms, calc):
        self.atoms = atoms
        self.calc = calc
        self.atoms.calc = self.calc

    def todict(self):
        return {'type': 'optimization',
                'optimizer': 'TurbomoleOptimizer'}

    def run(self, fmax=None, steps=None):
        if fmax is not None:
            self.calc.parameters['force convergence'] = fmax
            self.calc.verify_parameters()
        if steps is not None:
            self.calc.parameters['geometry optimization iterations'] = steps
            self.calc.verify_parameters()
        self.calc.calculate()
        self.atoms.positions[:] = self.calc.atoms.positions
        self.calc.parameters['task'] = 'energy'


class Turbomole(FileIOCalculator):

    """constants"""
    name = 'Turbomole'

    implemented_properties = ['energy', 'forces', 'dipole', 'free_energy',
                              'charges']

    available_functionals = [
        'slater-dirac-exchange', 's-vwn', 'vwn', 's-vwn_Gaussian', 'pwlda',
        'becke-exchange', 'b-lyp', 'b-vwn', 'lyp', 'b-p', 'pbe', 'tpss',
        'bh-lyp', 'b3-lyp', 'b3-lyp_Gaussian', 'pbe0', 'tpssh', 'lhf', 'oep',
        'b97-d', 'b2-plyp'
    ]
    tm_files = [
        'control', 'coord', 'basis', 'auxbasis', 'energy', 'gradient', 'mos',
        'alpha', 'beta', 'statistics', 'GEO_OPT_CONVERGED', 'GEO_OPT_FAILED',
        'not.converged', 'nextstep', 'hessapprox', 'job.last', 'job.start',
        'optinfo', 'statistics', 'converged', 'vibspectrum',
        'vib_normal_modes', 'hessian', 'dipgrad', 'dscf_problem', 'pc.txt',
        'pc_gradients.txt'
    ]
    tm_tmp_files = [
        'errvec', 'fock', 'oldfock', 'dens', 'ddens', 'diff_densmat',
        'diff_dft_density', 'diff_dft_oper', 'diff_fockmat', 'diis_errvec',
        'diis_oldfock'
    ]
    spec_names = {
        'default': 'default_parameters',
        'comment': 'parameter_comment',
        'updateable': 'parameter_updateable',
        'type': 'parameter_type',
        'key': 'parameter_key',
        'group': 'parameter_group',
        'units': 'parameter_units',
        'mapping': 'parameter_mapping',
        'non-define': 'parameter_no_define'
    }

    # flat dictionaries with parameters attributes
    default_parameters = {}
    parameter_comment = {}
    parameter_updateable = {}
    parameter_type = {}
    parameter_key = {}
    parameter_group = {}
    parameter_units = {}
    parameter_mapping = {}
    parameter_no_define = {}

    # nested dictionary with parameters attributes
    parameter_spec = {
        'automatic orbital shift': {
            'comment': None,
            'default': 0.1,
            'group': 'scforbitalshift',
            'key': 'automatic',
            'mapping': {
                'to_control': lambda a: a / Ha,
                'from_control': lambda a: a * Ha
            },
            'type': float,
            'units': 'eV',
            'updateable': True
        },
        'basis set definition': {
            'comment': 'used only in restart',
            'default': None,
            'group': 'basis',
            'key': None,
            'type': dict,
            'units': None,
            'updateable': False
        },
        'basis set name': {
            'comment': 'current default from module "define"',
            'default': 'def-SV(P)',
            'group': 'basis',
            'key': None,
            'type': str,
            'units': None,
            'updateable': False
        },
        'closed-shell orbital shift': {
            'comment': 'does not work with automatic',
            'default': None,
            'group': 'scforbitalshift',
            'key': 'closedshell',
            'mapping': {
                'to_control': lambda a: a / Ha,
                'from_control': lambda a: a * Ha
            },
            'type': float,
            'units': 'eV',
            'updateable': True
        },
        'damping adjustment step': {
            'comment': None,
            'default': None,
            'group': 'scfdamp',
            'key': 'step',
            'type': float,
            'units': None,
            'updateable': True
        },
        'density convergence': {
            'comment': None,
            'default': None,
            'group': 'denconv',
            'key': 'denconv',
            'mapping': {
                'to_control': lambda a: int(-log10(a)),
                'from_control': lambda a: 10**(-a)
            },
            'non-define': True,
            'type': float,
            'units': None,
            'updateable': True
        },
        'density functional': {
            'comment': None,
            'default': 'b-p',
            'group': 'dft',
            'key': 'functional',
            'type': str,
            'units': None,
            'updateable': True
        },
        'energy convergence': {
            'comment': 'jobex -energy <int>',
            'default': None,
            'group': None,
            'key': None,
            'mapping': {
                'to_control': lambda a: a / Ha,
                'from_control': lambda a: a * Ha
            },
            'type': float,
            'units': 'eV',
            'updateable': True
        },
        'fermi annealing factor': {
            'comment': None,
            'default': 0.95,
            'group': 'fermi',
            'key': 'tmfac',
            'type': float,
            'units': None,
            'updateable': True
        },
        'fermi final temperature': {
            'comment': None,
            'default': 300,
            'group': 'fermi',
            'key': 'tmend',
            'type': float,
            'units': 'Kelvin',
            'updateable': True
        },
        'fermi homo-lumo gap criterion': {
            'comment': None,
            'default': 0.1,
            'group': 'fermi',
            'key': 'hlcrt',
            'mapping': {
                'to_control': lambda a: a / Ha,
                'from_control': lambda a: a * Ha
            },
            'type': float,
            'units': 'eV',
            'updateable': True
        },
        'fermi initial temperature': {
            'comment': None,
            'default': 300,
            'group': 'fermi',
            'key': 'tmstrt',
            'type': float,
            'units': 'Kelvin',
            'updateable': True
        },
        'fermi stopping criterion': {
            'comment': None,
            'default': 0.001,
            'group': 'fermi',
            'key': 'stop',
            'mapping': {
                'to_control': lambda a: a / Ha,
                'from_control': lambda a: a * Ha
            },
            'type': float,
            'units': 'eV',
            'updateable': True
        },
        'force convergence': {
            'comment': 'jobex -gcart <int>',
            'default': None,
            'group': None,
            'key': None,
            'mapping': {
                'to_control': lambda a: a / Ha * Bohr,
                'from_control': lambda a: a * Ha / Bohr
            },
            'type': float,
            'units': 'eV/Angstrom',
            'updateable': True
        },
        'geometry optimization iterations': {
            'comment': 'jobex -c <int>',
            'default': None,
            'group': None,
            'key': None,
            'type': int,
            'units': None,
            'updateable': True
        },
        'grid size': {
            'comment': None,
            'default': 'm3',
            'group': 'dft',
            'key': 'gridsize',
            'type': str,
            'units': None,
            'updateable': True
        },
        'ground state': {
            'comment': 'only this is currently supported',
            'default': True,
            'group': None,
            'key': None,
            'type': bool,
            'units': None,
            'updateable': False
        },
        'initial damping': {
            'comment': None,
            'default': None,
            'group': 'scfdamp',
            'key': 'start',
            'type': float,
            'units': None,
            'updateable': True
        },
        'initial guess': {
            'comment': '"eht", "hcore" or {"use": "<path/to/control>"}',
            'default': 'eht',
            'group': None,
            'key': None,
            'type': None,
            'units': None,
            'updateable': False
        },
        'minimal damping': {
            'comment': None,
            'default': None,
            'group': 'scfdamp',
            'key': 'min',
            'type': float,
            'units': None,
            'updateable': True
        },
        'multiplicity': {
            'comment': None,
            'default': None,
            'group': None,
            'key': None,
            'type': int,
            'units': None,
            'updateable': False
        },
        'non-automatic orbital shift': {
            'comment': None,
            'default': False,
            'group': 'scforbitalshift',
            'key': 'noautomatic',
            'type': bool,
            'units': None,
            'updateable': True
        },
        'point group': {
            'comment': 'only c1 supported',
            'default': 'c1',
            'group': 'symmetry',
            'key': 'symmetry',
            'type': str,
            'units': None,
            'updateable': False
        },
        'ri memory': {
            'comment': None,
            'default': 1000,
            'group': 'ricore',
            'key': 'ricore',
            'type': int,
            'units': 'Megabyte',
            'updateable': True
        },
        'rohf': {
            'comment': 'used only in restart',
            'default': None,
            'group': None,
            'key': None,
            'type': bool,
            'units': None,
            'updateable': False
        },
        'scf energy convergence': {
            'comment': None,
            'default': None,
            'group': 'scfconv',
            'key': 'scfconv',
            'mapping': {
                'to_control': lambda a: int(floor(-log10(a / Ha))),
                'from_control': lambda a: 10**(-a) * Ha
            },
            'type': float,
            'units': 'eV',
            'updateable': True
        },
        'scf iterations': {
            'comment': None,
            'default': 60,
            'group': 'scfiterlimit',
            'key': 'scfiterlimit',
            'type': int,
            'units': None,
            'updateable': True
        },
        'task': {
            'comment': '"energy calculation" = "energy", '
                       '"gradient calculation" = "gradient", '
                       '"geometry optimization" = "optimize", '
                       '"normal mode analysis" = "frequencies"',
            'default': 'energy',
            'group': None,
            'key': None,
            'type': str,
            'units': None,
            'updateable': True
        },
        'title': {
            'comment': None,
            'default': '',
            'group': 'title',
            'key': 'title',
            'type': str,
            'units': None,
            'updateable': False
        },
        'total charge': {
            'comment': None,
            'default': 0,
            'group': None,
            'key': None,
            'type': int,
            'units': None,
            'updateable': False
        },
        'uhf': {
            'comment': None,
            'default': None,
            'group': 'uhf',
            'key': 'uhf',
            'type': bool,
            'units': None,
            'updateable': False
        },
        'use basis set library': {
            'comment': 'only true implemented',
            'default': True,
            'group': 'basis',
            'key': None,
            'type': bool,
            'units': None,
            'updateable': False
        },
        'use dft': {
            'comment': None,
            'default': True,
            'group': 'dft',
            'key': 'dft',
            'type': bool,
            'units': None,
            'updateable': False
        },
        'use fermi smearing': {
            'comment': None,
            'default': False,
            'group': 'fermi',
            'key': 'fermi',
            'type': bool,
            'units': None,
            'updateable': True
        },
        'use redundant internals': {
            'comment': None,
            'default': False,
            'group': 'redundant',
            'key': None,
            'type': bool,
            'units': None,
            'updateable': False
        },
        'use resolution of identity': {
            'comment': None,
            'default': False,
            'group': 'rij',
            'key': 'rij',
            'type': bool,
            'units': None,
            'updateable': False
        },
        'numerical hessian': {
            'comment': 'NumForce will be used if dictionary exists',
            'default': None,
            'group': None,
            'key': None,
            'type': dict,
            'units': None,
            'updateable': True
        },
        'esp fit': {
            'comment': 'ESP fit',
            'default': None,
            'group': 'esp_fit',
            'key': 'esp_fit',
            'type': str,
            'units': None,
            'updateable': True,
            'non-define': True
        }
    }

    # initialize attributes
    parameters = {}
    results = {}
    initialized = False
    pc_initialized = False
    converged = False
    updated = False
    update_energy = None
    update_forces = None
    update_geometry = None
    update_hessian = None
    atoms = None
    forces = None
    e_total = None
    dipole = None
    charges = None
    version = None
    runtime = None
    datetime = None
    hostname = None
    pcpot = None

    def __init__(self, label=None, calculate_energy='dscf',
                 calculate_forces='grad', post_HF=False, atoms=None,
                 restart=False, define_str=None, control_kdg=None,
                 control_input=None, **kwargs):

        FileIOCalculator.__init__(self)

        self.label = label
        self.calculate_energy = calculate_energy
        self.calculate_forces = calculate_forces
        self.post_HF = post_HF
        self.restart = restart
        self.define_str = define_str
        self.control_kdg = control_kdg
        self.control_input = control_input

        # construct flat dictionaries with parameter attributes
        for p in self.parameter_spec:
            for k in self.spec_names:
                if k in list(self.parameter_spec[p].keys()):
                    subdict = getattr(self, self.spec_names[k])
                    subdict.update({p: self.parameter_spec[p][k]})

        if self.restart:
            self._set_restart(kwargs)
        else:
            self.set_parameters(kwargs)
            self.verify_parameters()
            self.reset()

        if atoms is not None:
            atoms.set_calculator(self)
            self.set_atoms(atoms)

    def __getitem__(self, item):
        return getattr(self, item)

    def _set_restart(self, params_update):
        """constructs atoms, parameters and results from a previous
        calculation"""

        # read results, key parameters and non-key parameters
        self.read_restart()
        params_old = self.read_parameters()

        # filter out non-updateable parameters
        for p in list(params_update.keys()):
            if not self.parameter_updateable[p]:
                del params_update[p]
                warnings.warn('"' + p + '"' + ' cannot be changed')

        # update and verify parameters
        params_new = params_old.copy()
        params_new.update(params_update)
        self.set_parameters(params_new)
        self.verify_parameters()

        # if a define string is specified then run define
        if self.define_str:
            execute('define', input_str=self.define_str)

        # updates data groups in the control file
        if params_update or self.control_kdg or self.control_input:
            self._update_data_groups(params_old, params_update)

        self.initialized = True
        # more precise convergence tests are necessary to set these flags:
        self.update_energy = True
        self.update_forces = True
        self.update_geometry = True
        self.update_hessian = True

    def _update_data_groups(self, params_old, params_update):
        """updates data groups in the control file"""
        # construct a list of data groups to update
        grps = []
        for p in list(params_update.keys()):
            if self.parameter_group[p] is not None:
                grps.append(self.parameter_group[p])

        # construct a dictionary of data groups and update params
        dgs = {}
        for g in grps:
            dgs[g] = {}
            for p in self.parameter_key:
                if g == self.parameter_group[p]:
                    if self.parameter_group[p] == self.parameter_key[p]:
                        if p in list(params_update.keys()):
                            val = params_update[p]
                            pmap = list(self.parameter_mapping.keys())
                            if val is not None and p in pmap:
                                fun = self.parameter_mapping[p]['to_control']
                                val = fun(params_update[p])
                            dgs[g] = val
                    else:
                        if p in list(params_old.keys()):
                            val = params_old[p]
                            pmap = list(self.parameter_mapping.keys())
                            if val is not None and p in pmap:
                                fun = self.parameter_mapping[p]['to_control']
                                val = fun(params_old[p])
                            dgs[g][self.parameter_key[p]] = val
                        if p in list(params_update.keys()):
                            val = params_update[p]
                            pmap = list(self.parameter_mapping.keys())
                            if val is not None and p in pmap:
                                fun = self.parameter_mapping[p]['to_control']
                                val = fun(params_update[p])
                            dgs[g][self.parameter_key[p]] = val

        # write dgs dictionary to a data group
        for g in dgs:
            delete_data_group(g)
            if isinstance(dgs[g], dict):
                string = ''
                for key in list(dgs[g].keys()):
                    if dgs[g][key] is None:
                        continue
                    elif isinstance(dgs[g][key], bool):
                        if dgs[g][key]:
                            string += ' ' + key
                    else:
                        string += ' ' + key + '=' + str(dgs[g][key])
                add_data_group(g, string=string)
            else:
                if isinstance(dgs[g], bool):
                    if dgs[g]:
                        add_data_group(g, string='')
                else:
                    add_data_group(g, string=str(dgs[g]))

        self._set_post_define()

    def _set_post_define(self):
        """non-define keys, user-specified changes in the control file"""
        # process key parameters that are not written with define
        for p in list(self.parameters.keys()):
            if p in list(self.parameter_no_define.keys()):
                if self.parameter_no_define[p]:
                    if self.parameters[p]:
                        if p in list(self.parameter_mapping.keys()):
                            fun = self.parameter_mapping[p]['to_control']
                            val = fun(self.parameters[p])
                        else:
                            val = self.parameters[p]
                        delete_data_group(self.parameter_group[p])
                        add_data_group(self.parameter_group[p], str(val))
                    else:
                        delete_data_group(self.parameter_group[p])

        # delete user-specified data groups
        if self.control_kdg:
            for dg in self.control_kdg:
                delete_data_group(dg)

        # append user-defined input to control
        if self.control_input:
            for inp in self.control_input:
                add_data_group(inp, raw=True)

        # add point charges if pcpot defined:
        if self.pcpot:
            self.set_point_charges()

    def set_parameters(self, params):
        """loads the default parameters and updates with actual values"""
        self.parameters = self.default_parameters.copy()
        self.parameters.update(params)
        if self.parameters['use resolution of identity']:
            self.calculate_energy = 'ridft'
            self.calculate_forces = 'rdgrad'

    def verify_parameters(self):
        """detect wrong or not implemented parameters"""

        # kwargs parameters are ignored if user provides define_str
        if self.define_str is not None:
            assert isinstance(self.define_str, basestring)
            assert len(self.define_str) != 0
            return

        for par in self.parameters:
            assert par in self.parameter_spec, 'invalid parameter: ' + par

        if self.parameters['use dft']:
            func_list = [x.lower() for x in self.available_functionals]
            func = self.parameters['density functional']
            assert func.lower() in func_list, (
                'density functional not available / not supported'
            )

        assert self.parameters['multiplicity'], 'multiplicity not defined'

        if self.parameters['rohf']:
            raise NotImplementedError('ROHF not implemented')
        if self.parameters['initial guess'] not in ['eht', 'hcore']:
            if not (isinstance(self.parameters['initial guess'], dict) and
                    'use' in self.parameters['initial guess'].keys()):
                raise ValueError('Wrong input for initial guess')
        if not self.parameters['use basis set library']:
            raise NotImplementedError('Explicit basis set definition')
        if self.parameters['point group'] != 'c1':
            raise NotImplementedError('Point group not impemeneted')

    def reset(self):
        """removes all turbomole input, output and scratch files,
        and deletes results dict and the atoms object"""
        self.atoms = None
        self.results = {}
        self.results['calculation parameters'] = {}
        ase_files = [f for f in os.listdir('.') if f.startswith('ASE.TM.')]
        for f in self.tm_files + self.tm_tmp_files + ase_files:
            if os.path.exists(f):
                os.remove(f)
        self.initialized = False
        self.pc_initialized = False
        self.converged = False

    def set_atoms(self, atoms):
        """Create the self.atoms object and writes the coord file. If
        self.atoms exists a check for changes and an update of the atoms
        are performed. Note: Only positions changes are tracked in this
        version.
        """
        changes = self.check_state(atoms, tol=1e-13)
        if self.atoms == atoms or 'positions' not in changes:
            # print('two atoms obj are (almost) equal')
            if self.updated and os.path.isfile('coord'):
                self.updated = False
                a = read('coord').get_positions()
                if np.allclose(a, atoms.get_positions(), rtol=0, atol=1e-13):
                    return
            else:
                return

        changes = self.check_state(atoms, tol=1e-2)
        if 'positions' in changes:
            # print(two atoms obj are different')
            self.reset()
        else:
            # print('two atoms obj are slightly different')
            if self.parameters['use redundant internals']:
                self.reset()

        write('coord', atoms)
        self.atoms = atoms.copy()
        self.update_energy = True
        self.update_forces = True
        self.update_geometry = True
        self.update_hessian = True

    def get_define_str(self):
        """construct a define string from the parameters dictionary"""
        define_str_tpl = (
            '\n__title__\na coord\n__inter__\n'
            'bb all __basis_set__\n*\neht\ny\n__charge_str____occ_str__'
            '__single_atom_str____norb_str____dft_str____ri_str__'
            '__scfiterlimit____fermi_str____damp_str__q\n'
        )

        params = self.parameters

        if params['use redundant internals']:
            internals_str = 'ired\n*'
        else:
            internals_str = '*\nno'
        charge_str = str(params['total charge']) + '\n'

        if params['multiplicity'] == 1:
            if params['uhf']:
                occ_str = 'n\ns\n*\n'
            else:
                occ_str = 'y\n'
        elif params['multiplicity'] == 2:
            occ_str = 'y\n'
        elif params['multiplicity'] == 3:
            occ_str = 'n\nt\n*\n'
        else:
            unpaired = params['multiplicity'] - 1
            if params['use fermi smearing']:
                occ_str = 'n\nuf ' + str(unpaired) + '\n*\n'
            else:
                occ_str = 'n\nu ' + str(unpaired) + '\n*\n'

        if len(self.atoms) != 1:
            single_atom_str = ''
        else:
            single_atom_str = '\n'

        if params['multiplicity'] == 1:
            norb_str = ''
        else:
            norb_str = 'n\n'

        if params['use dft']:
            dft_str = 'dft\non\n*\n'
        else:
            dft_str = ''

        if params['density functional']:
            dft_str += 'dft\nfunc ' + params['density functional'] + '\n*\n'

        if params['grid size']:
            dft_str += 'dft\ngrid ' + params['grid size'] + '\n*\n'

        if params['use resolution of identity']:
            ri_str = 'ri\non\nm ' + str(params['ri memory']) + '\n*\n'
        else:
            ri_str = ''

        if params['scf iterations']:
            scfmaxiter = params['scf iterations']
            scfiter_str = 'scf\niter\n' + str(scfmaxiter) + '\n\n'
        else:
            scfiter_str = ''
        if params['scf energy convergence']:
            conv = floor(-log10(params['scf energy convergence'] / Ha))
            scfiter_str += 'scf\nconv\n' + str(int(conv)) + '\n\n'

        fermi_str = ''
        if params['use fermi smearing']:
            fermi_str = 'scf\nfermi\n'
            if params['fermi initial temperature']:
                par = str(params['fermi initial temperature'])
                fermi_str += '1\n' + par + '\n'
            if params['fermi final temperature']:
                par = str(params['fermi final temperature'])
                fermi_str += '2\n' + par + '\n'
            if params['fermi annealing factor']:
                par = str(params['fermi annealing factor'])
                fermi_str += '3\n' + par + '\n'
            if params['fermi homo-lumo gap criterion']:
                par = str(params['fermi homo-lumo gap criterion'])
                fermi_str += '4\n' + par + '\n'
            if params['fermi stopping criterion']:
                par = str(params['fermi stopping criterion'])
                fermi_str += '5\n' + par + '\n'
            fermi_str += '\n\n'

        damp_str = ''
        damp_keys = ('initial damping', 'damping adjustment step',
                     'minimal damping')
        damp_pars = [params[k] for k in damp_keys]
        if any(damp_pars):
            damp_str = 'scf\ndamp\n'
            for par in damp_pars:
                par_str = str(par) if par else ''
                damp_str += par_str + '\n'
            damp_str += '\n'

        define_str = define_str_tpl
        define_str = re.sub('__title__', params['title'], define_str)
        define_str = re.sub('__basis_set__', params['basis set name'],
                            define_str)
        define_str = re.sub('__charge_str__', charge_str, define_str)
        define_str = re.sub('__occ_str__', occ_str, define_str)
        define_str = re.sub('__norb_str__', norb_str, define_str)
        define_str = re.sub('__dft_str__', dft_str, define_str)
        define_str = re.sub('__ri_str__', ri_str, define_str)
        define_str = re.sub('__single_atom_str__', single_atom_str,
                            define_str)
        define_str = re.sub('__inter__', internals_str, define_str)
        define_str = re.sub('__scfiterlimit__', scfiter_str, define_str)
        define_str = re.sub('__fermi_str__', fermi_str, define_str)
        define_str = re.sub('__damp_str__', damp_str, define_str)

        return define_str

    def initialize(self):
        """prepare turbomole control file by running module 'define'"""
        if self.initialized:
            return
        self.verify_parameters()
        if not self.atoms:
            raise RuntimeError('atoms missing during initialization')
        if not os.path.isfile('coord'):
            raise IOError('file coord not found')

        if self.define_str is not None:
            define_str = self.define_str
        else:
            define_str = self.get_define_str()

        # run define
        execute('define', input_str=define_str)

        # process non-default initial guess
        iguess = self.parameters['initial guess']
        if isinstance(iguess, dict) and 'use' in iguess.keys():
            # "use" initial guess
            if self.parameters['multiplicity'] != 1 or self.parameters['uhf']:
                define_str = '\n\n\ny\nuse ' + iguess['use'] + '\nn\nn\nq\n'
            else:
                define_str = '\n\n\ny\nuse ' + iguess['use'] + '\nn\nq\n'
            execute('define', input_str=define_str)
        elif self.parameters['initial guess'] == 'hcore':
            # "hcore" initial guess
            if self.parameters['multiplicity'] != 1 or self.parameters['uhf']:
                delete_data_group('uhfmo_alpha')
                delete_data_group('uhfmo_beta')
                add_data_group('uhfmo_alpha', 'none file=alpha')
                add_data_group('uhfmo_beta', 'none file=beta')
            else:
                delete_data_group('scfmo')
                add_data_group('scfmo', 'none file=mos')

        self._set_post_define()

        self.initialized = True
        self.converged = False

    def calculation_required(self, atoms, properties):
        if self.atoms != atoms:
            return True
        for prop in properties:
            if prop == 'energy' and self.e_total is None:
                return True
            elif prop == 'forces' and self.forces is None:
                return True
        return False

    def calculate(self, atoms=None):
        """execute the requested job"""
        if atoms is None:
            atoms = self.atoms
        if self.parameters['task'] in ['energy', 'energy calculation']:
            self.get_potential_energy(atoms)
        if self.parameters['task'] in ['gradient', 'gradient calculation']:
            self.get_forces(atoms)
        if self.parameters['task'] in ['optimize', 'geometry optimization']:
            self.relax_geometry(atoms)
        if self.parameters['task'] in ['frequencies', 'normal mode analysis']:
            self.normal_mode_analysis(atoms)
        self.read_results()

    def relax_geometry(self, atoms=None):
        """execute geometry optimization with script jobex"""
        if atoms is None:
            atoms = self.atoms
        self.set_atoms(atoms)
        if self.converged and not self.update_geometry:
            return
        self.initialize()
        jobex_flags = ''
        if self.parameters['use resolution of identity']:
            jobex_flags += ' -ri'
        if self.parameters['force convergence']:
            par = self.parameters['force convergence']
            conv = floor(-log10(par / Ha * Bohr))
            jobex_flags += ' -gcart ' + str(int(conv))
        if self.parameters['energy convergence']:
            par = self.parameters['energy convergence']
            conv = floor(-log10(par / Ha))
            jobex_flags += ' -energy ' + str(int(conv))
        geom_iter = self.parameters['geometry optimization iterations']
        if geom_iter is not None:
            assert isinstance(geom_iter, int)
            jobex_flags += ' -c ' + str(geom_iter)
        self.converged = False
        execute('jobex' + jobex_flags)
        # check convergence
        self.converged = self.read_convergence()
        if self.converged:
            self.update_energy = False
            self.update_forces = False
            self.update_geometry = False
            self.update_hessian = True
        # read results
        new_struct = read('coord')
        atoms.set_positions(new_struct.get_positions())
        self.atoms = atoms.copy()
        self.read_energy()

    def normal_mode_analysis(self, atoms=None):
        """execute normal mode analysis with modules aoforce or NumForce"""
        from ase.constraints import FixAtoms
        if atoms is None:
            atoms = self.atoms
        self.set_atoms(atoms)
        self.initialize()
        if self.update_energy:
            self.get_potential_energy(atoms)
        if self.update_hessian:
            fixatoms = []
            for constr in atoms.constraints:
                if isinstance(constr, FixAtoms):
                    ckwargs = constr.todict()['kwargs']
                    if 'indices' in ckwargs.keys():
                        fixatoms.extend(ckwargs['indices'])
            if self.parameters['numerical hessian'] is None:
                if len(fixatoms) > 0:
                    define_str = '\n\ny\n'
                    for index in fixatoms:
                        define_str += 'm ' + str(index + 1) + ' 999.99999999\n'
                    define_str += '*\n*\nn\nq\n'
                    execute('define', input_str=define_str)
                    dg = read_data_group('atoms')
                    regex = r'(mass\s*=\s*)999.99999999'
                    dg = re.sub(regex, r'\g<1>9999999999.9', dg)
                    dg += '\n'
                    delete_data_group('atoms')
                    add_data_group(dg, raw=True)
                execute('aoforce')
            else:
                optstr = ''
                pdict = self.parameters['numerical hessian']
                if self.parameters['use resolution of identity']:
                    optstr += ' -ri'
                if len(fixatoms) > 0:
                    optstr += ' -frznuclei -central -c'
                if 'central' in pdict.keys():
                    optstr += ' -central'
                if 'delta' in pdict.keys():
                    optstr += ' -d ' + str(pdict['delta'] / Bohr)
                execute('NumForce' + optstr)
            self.update_hessian = False

    def read_restart(self):
        """read a previous calculation from control file"""
        self.atoms = read('coord')
        self.atoms.set_calculator(self)
        self.converged = self.read_convergence()
        read_methods = [
            self.read_energy,
            self.read_gradient,
            self.read_forces,
            self.read_basis_set,
            self.read_ecps,
            self.read_mos,
            self.read_occupation_numbers,
            self.read_dipole_moment,
            self.read_ssquare,
            self.read_hessian,
            self.read_vibrational_reduced_masses,
            self.read_normal_modes,
            self.read_vibrational_spectrum,
            self.read_charges,
            self.read_point_charges,
            self.read_run_parameters
        ]
        for method in read_methods:
            try:
                method()
            except ReadError as err:
                warnings.warn(err.args[0])

    def read_parameters(self):
        """read parameters from control file"""

        def parse_data_group(dg, dg_name):
            """parse a data group"""
            if len(dg) == 0:
                return None
            lsep = None
            ksep = None
            ndg = dg.replace('$' + dg_name, '').strip()
            if '\n' in ndg:
                lsep = '\n'
            if '=' in ndg:
                ksep = '='
            if not lsep and not ksep:
                return ndg
            result = {}
            lines = ndg.split(lsep)
            for line in lines:
                fields = line.strip().split(ksep)
                if len(fields) == 2:
                    result[fields[0]] = fields[1]
                elif len(fields) == 1:
                    result[fields[0]] = True
            return result

        params = {}
        pdgs = {}
        for p in self.parameter_group:
            if self.parameter_group[p] and self.parameter_key[p]:
                pdgs[p] = parse_data_group(
                    read_data_group(self.parameter_group[p]),
                    self.parameter_group[p]
                )

        for p in self.parameter_key:
            if self.parameter_key[p]:
                if self.parameter_key[p] == self.parameter_group[p]:
                    if pdgs[p] is None:
                        if self.parameter_type[p] is bool:
                            params[p] = False
                        else:
                            params[p] = None
                    else:
                        if self.parameter_type[p] is bool:
                            params[p] = True
                        else:
                            typ = self.parameter_type[p]
                            val = typ(pdgs[p])
                            mapping = self.parameter_mapping
                            if p in list(mapping.keys()):
                                fun = mapping[p]['from_control']
                                val = fun(val)
                            params[p] = val
                else:
                    if pdgs[p] is None:
                        params[p] = None
                    elif isinstance(pdgs[p], basestring):
                        if self.parameter_type[p] is bool:
                            params[p] = (pdgs[p] == self.parameter_key[p])
                    else:
                        if self.parameter_key[p] not in list(pdgs[p].keys()):
                            if self.parameter_type[p] is bool:
                                params[p] = False
                            else:
                                params[p] = None
                        else:
                            typ = self.parameter_type[p]
                            val = typ(pdgs[p][self.parameter_key[p]])
                            mapping = self.parameter_mapping
                            if p in list(mapping.keys()):
                                fun = mapping[p]['from_control']
                                val = fun(val)
                            params[p] = val

        # non-group or non-key parameters

        # per-element and per-atom basis sets not implemented in calculator
        basis_sets = set([bs['nickname'] for bs in self.results['basis set']])
        assert len(basis_sets) == 1
        params['basis set name'] = list(basis_sets)[0]
        params['basis set definition'] = self.results['basis set']

        # rohf, multiplicity and total charge
        orbs = self.results['molecular orbitals']
        params['rohf'] = (bool(len(read_data_group('rohf'))) or
                          bool(len(read_data_group('roothaan'))))
        core_charge = 0
        if self.results['ecps']:
            for ecp in self.results['ecps']:
                for symbol in self.atoms.get_chemical_symbols():
                    if symbol.lower() == ecp['element'].lower():
                        core_charge -= ecp['number of core electrons']
        if params['uhf']:
            alpha_occ = [o['occupancy'] for o in orbs if o['spin'] == 'alpha']
            beta_occ = [o['occupancy'] for o in orbs if o['spin'] == 'beta']
            spin = (np.sum(alpha_occ) - np.sum(beta_occ)) * 0.5
            params['multiplicity'] = int(2 * spin + 1)
            nuclear_charge = np.sum(self.atoms.numbers)
            electron_charge = -int(np.sum(alpha_occ) + np.sum(beta_occ))
            electron_charge += core_charge
            params['total charge'] = nuclear_charge + electron_charge
        elif not params['rohf']:  # restricted HF (closed shell)
            params['multiplicity'] = 1
            nuclear_charge = np.sum(self.atoms.numbers)
            electron_charge = -int(np.sum([o['occupancy'] for o in orbs]))
            electron_charge += core_charge
            params['total charge'] = nuclear_charge + electron_charge
        else:
            raise NotImplementedError('ROHF not implemented')

        # task-related parameters
        if os.path.exists('job.start'):
            with open('job.start', 'r') as log:
                lines = log.readlines()
            for line in lines:
                if 'CRITERION FOR TOTAL SCF-ENERGY' in line:
                    en = int(re.search(r'10\*{2}\(-(\d+)\)', line).group(1))
                    params['energy convergence'] = en
                if 'CRITERION FOR MAXIMUM NORM OF SCF-ENERGY GRADIENT' in line:
                    gr = int(re.search(r'10\*{2}\(-(\d+)\)', line).group(1))
                    params['force convergence'] = gr
                if 'AN OPTIMIZATION WITH MAX' in line:
                    cy = int(re.search(r'MAX. (\d+) CYCLES', line).group(1))
                    params['geometry optimization iterations'] = cy
        return params

    def read_convergence(self):
        """perform convergence checks"""
        if self.restart:
            if bool(len(read_data_group('restart'))):
                return False
            if bool(len(read_data_group('actual'))):
                return False
            if not bool(len(read_data_group('energy'))):
                return False
            if (os.path.exists('job.start') and
                os.path.exists('GEO_OPT_FAILED')):
                return False
            return True

        if self.parameters['task'] in ['optimize', 'geometry optimization']:
            if os.path.exists('GEO_OPT_CONVERGED'):
                return True
            elif os.path.exists('GEO_OPT_FAILED'):
                # check whether a failed scf convergence is the reason
                checkfiles = []
                for filename in os.listdir('.'):
                    if filename.startswith('job.'):
                        checkfiles.append(filename)
                for filename in checkfiles:
                    for line in open(filename):
                        if 'SCF FAILED TO CONVERGE' in line:
                            # scf did not converge in some jobex iteration
                            if filename == 'job.last':
                                raise RuntimeError('scf failed to converge')
                            else:
                                warnings.warn('scf failed to converge')
                warnings.warn('geometry optimization failed to converge')
                return False
            else:
                raise RuntimeError('error during geometry optimization')
        else:
            if os.path.isfile('dscf_problem'):
                raise RuntimeError('scf failed to converge')
            else:
                return True

    def read_results(self):
        """read all results and load them in the results entity"""
        self.read_energy()
        self.read_mos()
        self.read_basis_set()
        self.read_occupation_numbers()
        self.read_dipole_moment()
        self.read_ssquare()
        self.read_run_parameters()
        if self.parameters['task'] in ['gradient', 'optimize',
                                       'gradient calculation',
                                       'geometry optimization']:
            self.read_gradient()
            self.read_forces()
        if self.parameters['task'] in ['frequencies', 'normal mode analysis']:
            self.read_hessian()
            self.read_vibrational_reduced_masses()
            self.read_normal_modes()
            self.read_vibrational_spectrum()
        self.read_charges()

    def read_run_parameters(self):
        """read parameters set by define and not in self.parameters"""

        if 'calculation parameters' not in self.results.keys():
            self.results['calculation parameters'] = {}
        parameters = self.results['calculation parameters']
        dg = read_data_group('symmetry')
        parameters['point group'] = str(dg.split()[1])
        parameters['uhf'] = '$uhf' in read_data_group('uhf')
        # Gaussian function type
        gt = read_data_group('pople')
        if gt == '':
            parameters['gaussian type'] = 'spherical harmonic'
        else:
            gt = gt.split()[1]
            if gt == 'AO':
                parameters['gaussian type'] = 'spherical harmonic'
            elif gt == 'CAO':
                parameters['gaussian type'] = 'cartesian'
            else:
                parameters['gaussian type'] = None

        nvibro = read_data_group('nvibro')
        if nvibro:
            parameters['nuclear degrees of freedom'] = int(nvibro.split()[1])

    def read_energy(self):
        """Read energy from Turbomole energy file."""
        try:
            with open('energy', 'r') as enf:
                text = enf.read().lower()
        except IOError:
            raise ReadError('failed to read energy file')
        if text == '':
            raise ReadError('empty energy file')

        lines = iter(text.split('\n'))

        for line in lines:
            if line.startswith('$end'):
                break
            elif line.startswith('$'):
                pass
            else:
                energy_tmp = float(line.split()[1])
                if self.post_HF:
                    energy_tmp += float(line.split()[4])
        # update energy units
        self.e_total = energy_tmp * Ha
        self.results['total energy'] = self.e_total

    def read_forces(self):
        """Read Forces from Turbomole gradient file."""
        dg = read_data_group('grad')
        if len(dg) == 0:
            return
        file = open('gradient', 'r')
        lines = file.readlines()
        file.close()

        forces = np.array([[0, 0, 0]])

        nline = len(lines)
        iline = -1

        for i in range(nline):
            if 'cycle' in lines[i]:
                iline = i

        if iline < 0:
            raise RuntimeError('Please check TURBOMOLE gradients')

        # next line
        iline += len(self.atoms) + 1
        # $end line
        nline -= 1
        # read gradients
        for i in range(iline, nline):
            line = lines[i].replace('D', 'E')
            tmp = np.array([[float(f) for f in line.split()[0:3]]])
            forces = np.concatenate((forces, tmp))
        # Note the '-' sign for turbomole, to get forces
        self.forces = -np.delete(forces, np.s_[0:1], axis=0) * Ha / Bohr
        self.results['energy gradient'] = (-self.forces).tolist()

    def read_occupation_numbers(self):
        """read occupation numbers with module 'eiger' """
        if 'molecular orbitals' not in self.results.keys():
            return
        mos = self.results['molecular orbitals']
        args = ['eiger', '--all', '--pview']
        output = execute(args, error_test=False, stdout_tofile=False)
        lines = output.split('\n')
        for line in lines:
            regex = (
                r'^\s+(\d+)\.*\s+(\w*)\s+(\d+)\s+(\S+)'
                r'\s+(\d*\.*\d*)\s+([-+]?\d+\.\d*)'
            )
            match = re.search(regex, line)
            if match:
                orb_index = int(match.group(3))
                if match.group(2) == 'a':
                    spin = 'alpha'
                elif match.group(2) == 'b':
                    spin = 'beta'
                else:
                    spin = None
                ar_index = next(
                    index for (index, molecular_orbital) in enumerate(mos)
                    if (molecular_orbital['index'] == orb_index and
                        molecular_orbital['spin'] == spin)
                )
                mos[ar_index]['index by energy'] = int(match.group(1))
                irrep = str(match.group(4))
                mos[ar_index]['irreducible representation'] = irrep
                if match.group(5) != '':
                    mos[ar_index]['occupancy'] = float(match.group(5))
                else:
                    mos[ar_index]['occupancy'] = float(0)

    def read_mos(self):
        """read the molecular orbital coefficients and orbital energies
        from files mos, alpha and beta"""

        self.results['molecular orbitals'] = []
        mos = self.results['molecular orbitals']
        keywords = ['scfmo', 'uhfmo_alpha', 'uhfmo_beta']
        spin = [None, 'alpha', 'beta']

        for index, keyword in enumerate(keywords):
            flen = None
            mo = {}
            orbitals_coefficients_line = []
            mo_string = read_data_group(keyword)
            if mo_string == '':
                continue
            mo_string += '\n$end'
            lines = mo_string.split('\n')
            for line in lines:
                if re.match(r'^\s*#', line):
                    continue
                if 'eigenvalue' in line:
                    if len(orbitals_coefficients_line) != 0:
                        mo['eigenvector'] = orbitals_coefficients_line
                        mos.append(mo)
                        mo = {}
                        orbitals_coefficients_line = []
                    regex = (r'^\s*(\d+)\s+(\S+)\s+'
                             r'eigenvalue=([\+\-\d\.\w]+)\s')
                    match = re.search(regex, line)
                    mo['index'] = int(match.group(1))
                    mo['irreducible representation'] = str(match.group(2))
                    eig = float(re.sub('[dD]', 'E', match.group(3))) * Ha
                    mo['eigenvalue'] = eig
                    mo['spin'] = spin[index]
                    mo['degeneracy'] = 1
                    continue
                if keyword in line:
                    # e.g. format(4d20.14)
                    regex = r'format\(\d+[a-zA-Z](\d+)\.\d+\)'
                    match = re.search(regex, line)
                    if match:
                        flen = int(match.group(1))
                    if ('scfdump' in line or 'expanded' in line or
                        'scfconv' not in line):
                        self.converged = False
                    continue
                if '$end' in line:
                    if len(orbitals_coefficients_line) != 0:
                        mo['eigenvector'] = orbitals_coefficients_line
                        mos.append(mo)
                    break
                sfields = [line[i:i + flen]
                           for i in range(0, len(line), flen)]
                ffields = [float(f.replace('D', 'E').replace('d', 'E'))
                           for f in sfields]
                orbitals_coefficients_line += ffields

    def read_basis_set(self):
        """read the basis set"""
        self.results['basis set'] = []
        self.results['basis set formatted'] = {}
        bsf = read_data_group('basis')
        self.results['basis set formatted']['turbomole'] = bsf
        lines = bsf.split('\n')
        basis_set = {}
        functions = []
        function = {}
        primitives = []
        read_tag = False
        read_data = False
        for line in lines:
            if len(line.strip()) == 0:
                continue
            if '$basis' in line:
                continue
            if '$end' in line:
                break
            if re.match(r'^\s*#', line):
                continue
            if re.match(r'^\s*\*', line):
                if read_tag:
                    read_tag = False
                    read_data = True
                else:
                    if read_data:
                        # end primitives
                        function['primitive functions'] = primitives
                        function['number of primitives'] = len(primitives)
                        primitives = []
                        functions.append(function)
                        function = {}
                        # end contracted
                        basis_set['functions'] = functions
                        functions = []
                        self.results['basis set'].append(basis_set)
                        basis_set = {}
                        read_data = False
                    read_tag = True
                continue
            if read_tag:
                match = re.search(r'^\s*(\w+)\s+(.+)', line)
                if match:
                    basis_set['element'] = match.group(1)
                    basis_set['nickname'] = match.group(2)
                else:
                    raise RuntimeError('error reading basis set')
            else:
                match = re.search(r'^\s+(\d+)\s+(\w+)', line)
                if match:
                    if len(primitives) > 0:
                        # end primitives
                        function['primitive functions'] = primitives
                        function['number of primitives'] = len(primitives)
                        primitives = []
                        functions.append(function)
                        function = {}
                        # begin contracted
                    function['shell type'] = str(match.group(2))
                    continue
                regex = (
                    r'^\s*([-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)'
                    r'\s+([-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)'
                )
                match = re.search(regex, line)
                if match:
                    exponent = float(match.group(1))
                    coefficient = float(match.group(3))
                    primitives.append(
                        {'exponent': exponent, 'coefficient': coefficient}
                    )

    def read_ecps(self):
        """read the effective core potentials"""
        ecpf = read_data_group('ecp')
        if not bool(len(ecpf)):
            self.results['ecps'] = None
            self.results['ecps formatted'] = None
            return
        self.results['ecps'] = []
        self.results['ecps formatted'] = {}
        self.results['ecps formatted']['turbomole'] = ecpf
        lines = ecpf.split('\n')
        ecp = {}
        groups = []
        group = {}
        terms = []
        read_tag = False
        read_data = False
        for line in lines:
            if len(line.strip()) == 0:
                continue
            if '$ecp' in line:
                continue
            if '$end' in line:
                break
            if re.match(r'^\s*#', line):
                continue
            if re.match(r'^\s*\*', line):
                if read_tag:
                    read_tag = False
                    read_data = True
                else:
                    if read_data:
                        # end terms
                        group['terms'] = terms
                        group['number of terms'] = len(terms)
                        terms = []
                        groups.append(group)
                        group = {}
                        # end group
                        ecp['groups'] = groups
                        groups = []
                        self.results['ecps'].append(ecp)
                        ecp = {}
                        read_data = False
                    read_tag = True
                continue
            if read_tag:
                match = re.search(r'^\s*(\w+)\s+(.+)', line)
                if match:
                    ecp['element'] = match.group(1)
                    ecp['nickname'] = match.group(2)
                else:
                    raise RuntimeError('error reading ecp')
            else:
                regex = r'ncore\s*=\s*(\d+)\s+lmax\s*=\s*(\d+)'
                match = re.search(regex, line)
                if match:
                    ecp['number of core electrons'] = int(match.group(1))
                    ecp['maximum angular momentum number'] = \
                        int(match.group(2))
                    continue
                match = re.search(r'^(\w(\-\w)?)', line)
                if match:
                    if len(terms) > 0:
                        # end terms
                        group['terms'] = terms
                        group['number of terms'] = len(terms)
                        terms = []
                        groups.append(group)
                        group = {}
                        # begin group
                    group['title'] = str(match.group(1))
                    continue
                regex = (r'^\s*([-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s+'
                         r'(\d)\s+([-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)')
                match = re.search(regex, line)
                if match:
                    terms.append(
                        {
                            'coefficient': float(match.group(1)),
                            'power of r': float(match.group(3)),
                            'exponent': float(match.group(4))
                        }
                    )

    def read_gradient(self):
        """read all information in file 'gradient'"""
        from ase import Atom
        grad_string = read_data_group('grad')
        if len(grad_string) == 0:
            return
#       try to reuse ase:
#       structures = read('gradient', index=':')
        lines = grad_string.split('\n')
        history = []
        image = {}
        gradient = []
        atoms = Atoms()
        (cycle, energy, norm) = (None, None, None)
        for line in lines:
            # cycle lines
            regex = (
                r'^\s*cycle =\s*(\d+)\s+'
                r'SCF energy =\s*([-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)\s+'
                r'\|dE\/dxyz\| =\s*([-+]?[0-9]*\.?[0-9]+([eE][-+]?[0-9]+)?)'
            )
            match = re.search(regex, line)
            if match:
                if len(atoms):
                    image['optimization cycle'] = cycle
                    image['total energy'] = energy
                    image['gradient norm'] = norm
                    image['energy gradient'] = gradient
                    history.append(image)
                    image = {}
                    atoms = Atoms()
                    gradient = []
                cycle = int(match.group(1))
                energy = float(match.group(2)) * Ha
                norm = float(match.group(4)) * Ha / Bohr
                continue
            # coordinate lines
            regex = (
                r'^\s*([-+]?[0-9]*\.?[0-9]+([eEdD][-+]?[0-9]+)?)'
                r'\s+([-+]?[0-9]*\.?[0-9]+([eEdD][-+]?[0-9]+)?)'
                r'\s+([-+]?[0-9]*\.?[0-9]+([eEdD][-+]?[0-9]+)?)'
                r'\s+(\w+)'
            )
            match = re.search(regex, line)
            if match:
                x = float(match.group(1)) * Bohr
                y = float(match.group(3)) * Bohr
                z = float(match.group(5)) * Bohr
                symbol = str(match.group(7))
                atoms += Atom(symbol.capitalize(), (x, y, z))
                continue
            # gradient lines
            regex = (
                r'^\s*([-+]?[0-9]*\.?[0-9]+([eEdD][-+]?[0-9]+)?)'
                r'\s+([-+]?[0-9]*\.?[0-9]+([eEdD][-+]?[0-9]+)?)'
                r'\s+([-+]?[0-9]*\.?[0-9]+([eEdD][-+]?[0-9]+)?)'
            )
            match = re.search(regex, line)
            if match:
                gradx = float(match.group(1).replace('D', 'E')) * Ha / Bohr
                grady = float(match.group(3).replace('D', 'E')) * Ha / Bohr
                gradz = float(match.group(5).replace('D', 'E')) * Ha / Bohr
                gradient.append([gradx, grady, gradz])

        image['optimization cycle'] = cycle
        image['total energy'] = energy
        image['gradient norm'] = norm
        image['energy gradient'] = gradient
        history.append(image)
        self.results['geometry optimization history'] = history

    def read_hessian(self, noproj=False):
        """Read in the hessian matrix"""
        self.results['hessian matrix'] = {}
        self.results['hessian matrix']['array'] = []
        self.results['hessian matrix']['units'] = '?'
        self.results['hessian matrix']['projected'] = True
        self.results['hessian matrix']['mass weighted'] = True
        dg = read_data_group('nvibro')
        if len(dg) == 0:
            return
        nvibro = int(dg.split()[1])
        self.results['hessian matrix']['dimension'] = nvibro
        row = []
        key = 'hessian'
        if noproj:
            key = 'npr' + key
            self.results['hessian matrix']['projected'] = False
        lines = read_data_group(key).split('\n')
        for line in lines:
            if key in line:
                continue
            fields = line.split()
            row.extend(fields[2:len(fields)])
            if len(row) == nvibro:
                # check whether it is mass-weighted
                float_row = [float(element) for element in row]
                self.results['hessian matrix']['array'].append(float_row)
                row = []

    def read_normal_modes(self, noproj=False):
        """Read in vibrational normal modes"""
        self.results['normal modes'] = {}
        self.results['normal modes']['array'] = []
        self.results['normal modes']['projected'] = True
        self.results['normal modes']['mass weighted'] = True
        self.results['normal modes']['units'] = '?'
        dg = read_data_group('nvibro')
        if len(dg) == 0:
            return
        nvibro = int(dg.split()[1])
        self.results['normal modes']['dimension'] = nvibro
        row = []
        key = 'vibrational normal modes'
        if noproj:
            key = 'npr' + key
            self.results['normal modes']['projected'] = False
        lines = read_data_group(key).split('\n')
        for line in lines:
            if key in line:
                continue
            if '$end' in line:
                break
            fields = line.split()
            row.extend(fields[2:len(fields)])
            if len(row) == nvibro:
                # check whether it is mass-weighted
                float_row = [float(element) for element in row]
                self.results['normal modes']['array'].append(float_row)
                row = []

    def read_vibrational_reduced_masses(self):
        """Read vibrational reduced masses"""
        self.results['vibrational reduced masses'] = []
        dg = read_data_group('vibrational reduced masses')
        if len(dg) == 0:
            return
        lines = dg.split('\n')
        for line in lines:
            if '$vibrational' in line:
                continue
            if '$end' in line:
                break
            fields = [float(element) for element in line.split()]
            self.results['vibrational reduced masses'].extend(fields)

    def read_vibrational_spectrum(self, noproj=False):
        """Read the vibrational spectrum"""
        self.results['vibrational spectrum'] = []
        key = 'vibrational spectrum'
        if noproj:
            key = 'npr' + key
        lines = read_data_group(key).split('\n')
        for line in lines:
            dictionary = {}
            regex = (
                r'^\s+(\d+)\s+(\S*)\s+([-+]?\d+\.\d*)'
                r'\s+(\d+\.\d*)\s+(\S+)\s+(\S+)'
            )
            match = re.search(regex, line)
            if match:
                dictionary['mode number'] = int(match.group(1))
                dictionary['irreducible representation'] = str(match.group(2))
                dictionary['frequency'] = {
                    'units': 'cm^-1',
                    'value': float(match.group(3))
                }
                dictionary['infrared intensity'] = {
                    'units': 'km/mol',
                    'value': float(match.group(4))
                }

                if match.group(5) == 'YES':
                    dictionary['infrared active'] = True
                elif match.group(5) == 'NO':
                    dictionary['infrared active'] = False
                else:
                    dictionary['infrared active'] = None

                if match.group(6) == 'YES':
                    dictionary['Raman active'] = True
                elif match.group(6) == 'NO':
                    dictionary['Raman active'] = False
                else:
                    dictionary['Raman active'] = None

                self.results['vibrational spectrum'].append(dictionary)

    def read_ssquare(self):
        """Read the expectation value of S^2 operator"""
        s2_string = read_data_group('ssquare from dscf')
        if s2_string == '':
            return
        string = s2_string.split('\n')[1]
        ssquare = float(re.search(r'^\s*(\d+\.*\d*)', string).group(1))
        self.results['ssquare from scf calculation'] = ssquare

    def read_dipole_moment(self):
        """Read the dipole moment"""
        dip_string = read_data_group('dipole')
        if dip_string == '':
            return
        lines = dip_string.split('\n')
        for line in lines:
            regex = (
                r'^\s+x\s+([-+]?\d+\.\d*)\s+y\s+([-+]?\d+\.\d*)'
                r'\s+z\s+([-+]?\d+\.\d*)\s+a\.u\.'
            )
            match = re.search(regex, line)
            if match:
                dip_vec = [float(match.group(c)) for c in range(1, 4)]
            regex = r'^\s+\| dipole \| =\s+(\d+\.*\d*)\s+debye'
            match = re.search(regex, line)
            if match:
                dip_abs_val = float(match.group(1))
        self.results['electric dipole moment'] = {}
        self.results['electric dipole moment']['vector'] = {
            'array': dip_vec,
            'units': 'a.u.'
        }
        self.results['electric dipole moment']['absolute value'] = {
            'value': dip_abs_val,
            'units': 'Debye'
        }
        self.dipole = np.array(dip_vec) * Bohr

    def read_version(self):
        """read the version from the tm output if stored in a file"""
        versions = read_output(r'TURBOMOLE\s+V(\d+\.\d+)\s+')
        if len(set(versions)) > 1:
            warnings.warn('different turbomole versions detected')
            self.version = list(set(versions))
        elif len(versions) == 0:
            warnings.warn('no turbomole version detected')
            self.version = None
        else:
            self.version = versions[0]

    def read_datetime(self):
        """read the datetime of the most recent calculation
        from the tm output if stored in a file
        """
        datetimes = read_output(
            r'(\d{4}-[01]\d-[0-3]\d([T\s][0-2]\d:[0-5]'
            r'\d:[0-5]\d\.\d+)?([+-][0-2]\d:[0-5]\d|Z)?)')
        if len(datetimes) == 0:
            warnings.warn('no turbomole datetime detected')
            self.datetime = None
        else:
            # take the most recent time stamp
            self.datetime = sorted(datetimes, reverse=True)[0]

    def read_runtime(self):
        """read the total runtime of calculations"""
        hits = read_output(r'total wall-time\s+:\s+(\d+.\d+)\s+seconds')
        if len(hits) == 0:
            warnings.warn('no turbomole runtimes detected')
            self.runtime = None
        else:
            self.runtime = np.sum([float(a) for a in hits])

    def read_hostname(self):
        """read the hostname of the computer on which the calc has run"""
        hostnames = read_output(r'hostname is\s+(.+)')
        if len(set(hostnames)) > 1:
            warnings.warn('runs on different hosts detected')
            self.hostname = list(set(hostnames))
        else:
            self.hostname = hostnames[0]

    def get_optimizer(self, atoms, trajectory=None, logfile=None):
        """returns a TurbomoleOptimizer object"""
        self.parameters['task'] = 'optimize'
        self.verify_parameters()
        return TurbomoleOptimizer(atoms, self)

    def get_results(self):
        """returns the results dictionary"""
        return self.results

    def get_potential_energy(self, atoms, force_consistent=True):
        # update atoms
        self.updated = self.e_total is None
        self.set_atoms(atoms)
        self.initialize()
        # if update of energy is necessary
        if self.update_energy:
            # calculate energy
            execute(self.calculate_energy)
            # check convergence
            self.converged = self.read_convergence()
            if not self.converged:
                return None
            # read energy
            self.read_energy()

        self.update_energy = False
        return self.e_total

    def get_forces(self, atoms):
        # update atoms
        self.updated = self.forces is None
        self.set_atoms(atoms)
        # complete energy calculations
        if self.update_energy:
            self.get_potential_energy(atoms)
        # if update of forces is necessary
        if self.update_forces:
            # calculate forces
            execute(self.calculate_forces)
            # read forces
            self.read_forces()

        self.update_forces = False
        return self.forces.copy()

    def get_dipole_moment(self, atoms):
        if self.update_energy:
            self.get_potential_energy(atoms)
        self.read_dipole_moment()
        return self.dipole

    def get_property(self, name, atoms=None, allow_calculation=True):
        """return the value of a property"""

        if name not in self.implemented_properties:
            # an ugly work around; the caller should test the raised error
            # if name in ['magmom', 'magmoms', 'charges', 'stress']:
            # return None
            raise PropertyNotImplementedError(name)

        if atoms is None:
            atoms = self.atoms.copy()

        persist_property = {
            'energy': 'e_total',
            'forces': 'forces',
            'dipole': 'dipole',
            'free_energy': 'e_total',
            'charges': 'charges'
        }
        property_getter = {
            'energy': self.get_potential_energy,
            'forces': self.get_forces,
            'dipole': self.get_dipole_moment,
            'free_energy': self.get_potential_energy,
            'charges': self.get_charges
        }
        getter_args = {
            'energy': [atoms],
            'forces': [atoms],
            'dipole': [atoms],
            'free_energy': [atoms, True],
            'charges': [atoms]
        }

        if allow_calculation:
            result = property_getter[name](*getter_args[name])
        else:
            if hasattr(self, persist_property[name]):
                result = getattr(self, persist_property[name])
            else:
                result = None

        if isinstance(result, np.ndarray):
            result = result.copy()
        return result

    def get_charges(self, atoms):
        """return partial charges on atoms from an ESP fit"""
        if self.charges is None:
            self.calculate(atoms)
            self.read_charges()
        return self.charges

    def read_charges(self):
        """read partial charges on atoms from an ESP fit"""
        epsfit_defined = ('esp fit' in self.parameters and
                          self.parameters['esp fit'] is not None)
        if epsfit_defined or len(read_data_group('esp_fit')) > 0:
            filename = 'ASE.TM.' + self.calculate_energy + '.out'
            with open(filename, 'r') as infile:
                lines = infile.readlines()
            oklines = None
            for n, line in enumerate(lines):
                if 'atom  radius/au   charge' in line:
                    oklines = lines[n + 1:n + len(self.atoms) + 1]
            if oklines is not None:
                qm_charges = [float(line.split()[3]) for line in oklines]
                self.charges = np.array(qm_charges)

    def get_forces_on_point_charges(self):
        """return forces acting on point charges"""
        self.get_forces(self.atoms)
        lines = read_data_group('point_charge_gradients').split('\n')[1:]
        forces = []
        for line in lines:
            linef = line.strip().replace('D', 'E')
            forces.append([float(x) for x in linef.split()])
        # Note the '-' sign for turbomole, to get forces
        return -np.array(forces) * Ha / Bohr

    def set_point_charges(self, pcpot=None):
        """write external point charges to control"""
        if pcpot is not None and pcpot != self.pcpot:
            self.pcpot = pcpot
        if self.pcpot.mmcharges is None or self.pcpot.mmpositions is None:
            raise RuntimeError('external point charges not defined')

        if not self.pc_initialized:
            if len(read_data_group('point_charges')) == 0:
                add_data_group('point_charges', 'file=pc.txt')
            if len(read_data_group('point_charge_gradients')) == 0:
                add_data_group(
                    'point_charge_gradients',
                    'file=pc_gradients.txt'
                )
            drvopt = read_data_group('drvopt')
            if 'point charges' not in drvopt:
                drvopt += '\n   point charges\n'
                delete_data_group('drvopt')
                add_data_group(drvopt, raw=True)
            self.pc_initialized = True

        if self.pcpot.updated:
            with open('pc.txt', 'w') as pcfile:
                pcfile.write('$point_charges nocheck list\n')
                for (x, y, z), charge in zip(
                        self.pcpot.mmpositions, self.pcpot.mmcharges):
                    pcfile.write('%20.14f  %20.14f  %20.14f  %20.14f\n'
                                 % (x / Bohr, y / Bohr, z / Bohr, charge))
                pcfile.write('$end \n')
            self.pcpot.updated = False

    def read_point_charges(self):
        """read point charges from previous calculation"""
        pcs = read_data_group('point_charges')
        if len(pcs) > 0:
            lines = pcs.split('\n')[1:]
            (charges, positions) = ([], [])
            for line in lines:
                columns = [float(col) for col in line.strip().split()]
                positions.append([col * Bohr for col in columns[0:3]])
                charges.append(columns[3])
            self.pcpot = PointChargePotential(charges, positions)

    def embed(self, charges=None, positions=None):
        """embed atoms in an array of point-charges; function used in
            qmmm calculations."""
        self.pcpot = PointChargePotential(charges, positions)
        return self.pcpot


class PointChargePotential:
    """Point-charge potential for Turbomole"""
    def __init__(self, mmcharges, mmpositions=None):
        self.mmcharges = mmcharges
        self.mmpositions = mmpositions
        self.mmforces = None
        self.updated = True

    def set_positions(self, mmpositions):
        """set the positions of point charges"""
        self.mmpositions = mmpositions
        self.updated = True

    def set_charges(self, mmcharges):
        """set the values of point charges"""
        self.mmcharges = mmcharges
        self.updated = True

    def get_forces(self, calc):
        """forces acting on point charges"""
        self.mmforces = calc.get_forces_on_point_charges()
        return self.mmforces



import sys
import numpy as np

from ase.vibrations.franck_condon import FranckCondonOverlap, FranckCondonRecursive
from math import factorial


def equal(x, y, tolerance=0, fail=True, msg=''):
    """Compare x and y."""

    if not np.isfinite(x - y).any() or (np.abs(x - y) > tolerance).any():
        msg = (msg + '%s != %s (error: |%s| > %.9g)' %
               (x, y, x - y, tolerance))
        if fail:
            raise AssertionError(msg)
        else:
            sys.stderr.write('WARNING: %s\n' % msg)

# FCOverlap

fco = FranckCondonOverlap()
fcr = FranckCondonRecursive()

# check factorial
assert(fco.factorial(8) == factorial(8))
# the second test is useful according to the implementation
assert(fco.factorial(5) == factorial(5))
assert(fco.factorial.inv(5) == 1. / factorial(5))

# check T=0 and n=0 equality
S = np.array([1, 2.1, 34])
m = 5
assert(((fco.directT0(m, S) - fco.direct(0, m, S)) / fco.directT0(m, S) <
        1e-15).all())

# check symmetry
S = 2
n = 3
assert(fco.direct(n, m, S) == fco.direct(m, n, S))

# ---------------------------
# specials
S = np.array([0, 1.5])
delta = np.sqrt(2 * S)
for m in [2, 7]:
    equal(fco.direct0mm1(m, S)**2,
          fco.direct(1, m, S) * fco.direct(m, 0, S), 1.e-17)
    equal(fco.direct0mm1(m, S), fcr.ov0mm1(m, delta), 1.e-15)
    equal(fcr.ov0mm1(m, delta),
          fcr.ov0m(m, delta) * fcr.ov1m(m, delta), 1.e-15)
    equal(fcr.ov0mm1(m, -delta), fcr.direct0mm1(m, -delta), 1.e-15)
    equal(fcr.ov0mm1(m, delta), - fcr.direct0mm1(m, -delta), 1.e-15)

    equal(fco.direct0mm2(m, S)**2,
          fco.direct(2, m, S) * fco.direct(m, 0, S), 1.e-17)
    equal(fco.direct0mm2(m, S), fcr.ov0mm2(m, delta), 1.e-15)
    equal(fcr.ov0mm2(m, delta),
          fcr.ov0m(m, delta) * fcr.ov2m(m, delta), 1.e-15)
    equal(fco.direct0mm2(m, S), fcr.direct0mm2(m, delta), 1.e-15)

    equal(fcr.direct0mm3(m, delta),
          fcr.ov0m(m, delta) * fcr.ov3m(m, delta), 1.e-15)
    
    equal(fcr.ov1mm2(m, delta),
          fcr.ov1m(m, delta) * fcr.ov2m(m, delta), 1.e-15)
    equal(fcr.direct1mm2(m, delta), fcr.ov1mm2(m, delta), 1.e-15)




from ase.calculators.calculator import Calculator

calc = Calculator()
assert calc.directory == '.'
assert calc.prefix is None
assert calc.label is None

calc.label = 'dir/pref'
assert calc.directory == 'dir'
assert calc.prefix == 'pref'
assert calc.label == 'dir/pref'

calc.label = 'dir2/'
assert calc.directory == 'dir2'
assert calc.prefix is None
assert calc.label == 'dir2/'

calc.label = 'hello'
assert calc.directory == '.'
assert calc.prefix == 'hello'
assert calc.label == 'hello'

calc.label = None
assert calc.label is None
assert calc.prefix is None
assert calc.directory == '.'



# flake8: noqa
"""
The following contains a database of small molecules

Data for the G2/97 database are from
Raghavachari, Redfern, and Pople, J. Chem. Phys. Vol. 106, 1063 (1997).
See http://www.cse.anl.gov/Catalysis_and_Energy_Conversion/Computational_Thermochemistry.shtml for the original files.

All numbers are experimental values, except for coordinates, which are
MP2(full)/6-31G(d) optimized geometries (from http://www.cse.anl.gov/OldCHMwebsiteContent/compmat/G2-97.htm)

Atomic species:
ref: Curtiss et al. JCP 106, 1063 (1997).
'Enthalpy' is the experimental enthalpies of formation at 0K
'thermal correction' is the thermal corrections H(298)-H(0)

Molecular species:
ref: Staroverov et al. JCP 119, 12129 (2003)
'Enthalpy' is the experimental enthalpies of formation at 298K
'ZPE' is the zero-point energies
'thermal correction' is the thermal enthalpy corrections H(298K) - H_exp(0K)
ZPE and thermal corrections are estimated from B3LYP geometries and vibrations.

For details about G2-1 and G2-2 sets see doi:10.1063/1.477422.

Experimental ionization potentials are from http://srdata.nist.gov/cccbdb/
Information presented on these pages is considered public information
and may be distributed or copied http://www.nist.gov/public_affairs/disclaimer.cfm

"""

from ase.symbols import string2symbols

atom_names = ['H','Li','Be','C','N','O','F','Na','Si','P','S','Cl']

molecule_names = ['LiH','BeH','CH','CH2_s3B1d','CH2_s1A1d','CH3','CH4','NH','NH2','NH3','OH','H2O','HF','SiH2_s1A1d','SiH2_s3B1d','SiH3','SiH4','PH2','PH3','SH2','HCl','Li2','LiF','C2H2','C2H4','C2H6','CN','HCN','CO','HCO','H2CO','CH3OH','N2','N2H4','NO','O2','H2O2','F2','CO2','Na2','Si2','P2','S2','Cl2','NaCl','SiO','CS','SO','ClO','ClF','Si2H6','CH3Cl','CH3SH','HOCl','SO2']

data = {
 'Be': {'CAS No.': 7440417,
        'charges': None,
        'database': 'G2-1',
        'description': 'Be atom',
        'enthalpy': 76.48,
        'ionization energy': 9.32,
        'magmoms': None,
        'name': 'Beryllium',
        'positions': [[0.0, 0.0, 0.0]],
        'symbols': 'Be',
        'thermal correction': 0.46},
 'BeH': {'CAS No.': 13597972,
         'ZPE': 2.9073,
         'charges': None,
         'database': 'G2-1',
         'description': 'Beryllium hydride (BeH), D*h symm.',
         'enthalpy': 81.7,
         'ionization energy': 8.21,
         'magmoms': [0.8, 0.2],
         'name': 'BeH (beryllium monohydride)',
         'positions': [[0.0, 0.0, 0.269654], [0.0, 0.0, -1.078616]],
         'symbols': 'BeH',
         'thermal correction': 2.0739},
 'C': {'CAS No.': 7440440,
       'charges': None,
       'database': 'G2-1',
       'description': 'C atom',
       'enthalpy': 169.98,
       'ionization energy': 11.26,
       'magmoms': [2.0],
       'name': 'Carbon',
       'positions': [[0.0, 0.0, 0.0]],
       'symbols': 'C',
       'thermal correction': 0.25},
 'C2H2': {'CAS No.': 74862,
          'ZPE': 16.6001,
          'charges': None,
          'database': 'G2-1',
          'description': 'Acetylene (C2H2), D*h symm.',
          'enthalpy': 54.2,
          'ionization energy': 11.4,
          'magmoms': None,
          'name': 'C_2H_2',
          'positions': [[0.0, 0.0, 0.60808],
                        [0.0, 0.0, -0.60808],
                        [0.0, 0.0, -1.67399],
                        [0.0, 0.0, 1.67399]],
          'symbols': 'CCHH',
          'thermal correction': 2.4228,
          'vertical ionization energy': 11.49},
 'C2H4': {'CAS No.': 74851,
          'ZPE': 31.5267,
          'charges': None,
          'database': 'G2-1',
          'description': 'Ethylene (H2C=CH2), D2h symm.',
          'enthalpy': 12.5,
          'ionization energy': 11.4,
          'magmoms': None,
          'name': 'C_2H_4',
          'positions': [[0.0, 0.0, 0.66748],
                        [0.0, 0.0, -0.66748],
                        [0.0, 0.922832, 1.237695],
                        [0.0, -0.922832, 1.237695],
                        [0.0, 0.922832, -1.237695],
                        [0.0, -0.922832, -1.237695]],
          'symbols': 'CCHHHH',
          'thermal correction': 2.51,
          'vertical ionization energy': 11.49},
 'C2H6': {'CAS No.': 74840,
          'ZPE': 46.095,
          'charges': None,
          'database': 'G2-1',
          'description': 'Ethane (H3C-CH3), D3d symm.',
          'enthalpy': -20.1,
          'magmoms': None,
          'name': 'C_2H_6',
          'positions': [[0.0, 0.0, 0.762209],
                        [0.0, 0.0, -0.762209],
                        [0.0, 1.018957, 1.157229],
                        [-0.882443, -0.509479, 1.157229],
                        [0.882443, -0.509479, 1.157229],
                        [0.0, -1.018957, -1.157229],
                        [-0.882443, 0.509479, -1.157229],
                        [0.882443, 0.509479, -1.157229]],
          'symbols': 'CCHHHHHH',
          'thermal correction': 2.7912},
 'CH': {'CAS No.': 3315375,
        'ZPE': 3.9659,
        'charges': None,
        'database': 'G2-1',
        'description': 'CH radical. Doublet, C*v symm.',
        'enthalpy': 142.5,
        'ionization energy': 10.64,
        'magmoms': [1.0, 0.0],
        'name': 'CH(Methylidyne)',
        'positions': [[0.0, 0.0, 0.160074], [0.0, 0.0, -0.960446]],
        'symbols': 'CH',
        'thermal correction': 2.0739},
 'CH2_s1A1d': {'CAS No.': 2465567,
               'ZPE': 10.2422,
               'charges': None,
               'database': 'G2-1',
               'description': 'Singlet methylene (CH2), C2v symm, 1-A1.',
               'enthalpy': 102.8,
               'magmoms': None,
               'name': 'CH_2 (^1A_1)',
               'positions': [[0.0, 0.0, 0.174343],
                             [0.0, 0.862232, -0.523029],
                             [0.0, -0.862232, -0.523029]],
               'symbols': 'CHH',
               'thermal correction': 2.3745},
 'CH2_s3B1d': {'CAS No.': 2465567,
               'ZPE': 10.6953,
               'charges': None,
               'database': 'G2-1',
               'description': 'Triplet methylene (CH2), C2v symm, 3-B1.',
               'enthalpy': 93.7,
               'ionization energy': 10.4,
               'magmoms': [2.0, 0.0, 0.0],
               'name': 'CH_2 (^3B_1)',
               'positions': [[0.0, 0.0, 0.110381],
                             [0.0, 0.982622, -0.331142],
                             [0.0, -0.982622, -0.331142]],
               'symbols': 'CHH',
               'thermal correction': 2.3877},
 'CH3': {'CAS No.': 2229074,
         'ZPE': 18.3383,
         'charges': None,
         'database': 'G2-1',
         'description': 'Methyl radical (CH3), D3h symm.',
         'enthalpy': 35.0,
         'ionization energy': 9.84,
         'magmoms': [1.0, 0.0, 0.0, 0.0],
         'name': 'CH_3',
         'positions': [[0.0, 0.0, 0.0],
                       [0.0, 1.07841, 0.0],
                       [0.93393, -0.539205, 0.0],
                       [-0.93393, -0.539205, 0.0]],
         'symbols': 'CHHH',
         'thermal correction': 2.5383},
 'CH3Cl': {'CAS No.': 74873,
           'ZPE': 23.3013,
           'charges': None,
           'database': 'G2-1',
           'description': 'Methyl chloride (CH3Cl), C3v symm.',
           'enthalpy': -19.6,
           'ionization energy': 11.26,
           'magmoms': None,
           'name': 'CH_3Cl',
           'positions': [[0.0, 0.0, -1.121389],
                         [0.0, 0.0, 0.655951],
                         [0.0, 1.029318, -1.47428],
                         [0.891415, -0.514659, -1.47428],
                         [-0.891415, -0.514659, -1.47428]],
           'symbols': 'CClHHH',
           'thermal correction': 2.4956,
           'vertical ionization energy': 11.29},
 'CH3OH': {'CAS No.': 67561,
           'ZPE': 31.6635,
           'charges': None,
           'database': 'G2-1',
           'description': 'Methanol (CH3-OH), Cs symm.',
           'enthalpy': -48.0,
           'ionization energy': 10.84,
           'magmoms': None,
           'name': 'H_3COH',
           'positions': [[-0.047131, 0.664389, 0.0],
                         [-0.047131, -0.758551, 0.0],
                         [-1.092995, 0.969785, 0.0],
                         [0.878534, -1.048458, 0.0],
                         [0.437145, 1.080376, 0.891772],
                         [0.437145, 1.080376, -0.891772]],
           'symbols': 'COHHHH',
           'thermal correction': 2.6832,
           'vertical ionization energy': 10.96},
 'CH3SH': {'CAS No.': 74931,
           'ZPE': 28.3973,
           'charges': None,
           'database': 'G2-1',
           'description': 'Methanethiol (H3C-SH), Staggered, Cs symm.',
           'enthalpy': -5.5,
           'ionization energy': 9.44,
           'magmoms': None,
           'name': 'H_3CSH',
           'positions': [[-0.047953, 1.149519, 0.0],
                         [-0.047953, -0.664856, 0.0],
                         [1.283076, -0.823249, 0.0],
                         [-1.092601, 1.461428, 0.0],
                         [0.432249, 1.551207, 0.892259],
                         [0.432249, 1.551207, -0.892259]],
           'symbols': 'CSHHHH',
           'thermal correction': 2.869,
           'vertical ionization energy': 9.44},
 'CH4': {'CAS No.': 74828,
         'ZPE': 27.6744,
         'charges': None,
         'database': 'G2-1',
         'description': 'Methane (CH4), Td symm.',
         'enthalpy': -17.9,
         'ionization energy': 12.64,
         'magmoms': None,
         'name': 'CH_4',
         'positions': [[0.0, 0.0, 0.0],
                       [0.629118, 0.629118, 0.629118],
                       [-0.629118, -0.629118, 0.629118],
                       [0.629118, -0.629118, -0.629118],
                       [-0.629118, 0.629118, -0.629118]],
         'symbols': 'CHHHH',
         'thermal correction': 2.3939,
         'vertical ionization energy': 13.6},
 'CN': {'CAS No.': 2074875,
        'ZPE': 3.0183,
        'charges': None,
        'database': 'G2-1',
        'description': 'Cyano radical (CN), C*v symm, 2-Sigma+.',
        'enthalpy': 104.9,
        'ionization energy': 13.6,
        'magmoms': [1.0, 0.0],
        'name': 'CN (Cyano radical)',
        'positions': [[0.0, 0.0, -0.611046], [0.0, 0.0, 0.523753]],
        'symbols': 'CN',
        'thermal correction': 2.0739},
 'CO': {'CAS No.': 630080,
        'ZPE': 3.1062,
        'charges': None,
        'database': 'G2-1',
        'description': 'Carbon monoxide (CO), C*v symm.',
        'enthalpy': -26.4,
        'ionization energy': 14.01,
        'magmoms': None,
        'name': 'CO',
        'positions': [[0.0, 0.0, 0.493003], [0.0, 0.0, -0.657337]],
        'symbols': 'OC',
        'thermal correction': 2.0739,
        'vertical ionization energy': 14.01},
 'CO2': {'CAS No.': 124389,
         'ZPE': 7.313,
         'charges': None,
         'database': 'G2-1',
         'description': 'Carbon dioxide (CO2), D*h symm.',
         'enthalpy': -94.1,
         'ionization energy': 13.78,
         'magmoms': None,
         'name': 'CO_2',
         'positions': [[0.0, 0.0, 0.0],
                       [0.0, 0.0, 1.178658],
                       [0.0, 0.0, -1.178658]],
         'symbols': 'COO',
         'thermal correction': 2.2321,
         'vertical ionization energy': 13.78},
 'CS': {'CAS No.': 2944050,
        'ZPE': 1.8242,
        'charges': None,
        'database': 'G2-1',
        'description': 'Carbon monosulfide (CS), C*v symm.',
        'enthalpy': 66.9,
        'ionization energy': 11.33,
        'magmoms': None,
        'name': 'SC',
        'positions': [[0.0, 0.0, -1.123382], [0.0, 0.0, 0.421268]],
        'symbols': 'CS',
        'thermal correction': 2.0814},
 'Cl': {'CAS No.': 22537151,
        'charges': None,
        'database': 'G2-1',
        'description': 'Cl atom',
        'enthalpy': 28.59,
        'ionization energy': 12.97,
        'magmoms': [1.0],
        'name': 'Chlorine',
        'positions': [[0.0, 0.0, 0.0]],
        'symbols': 'Cl',
        'thermal correction': 1.1},
 'Cl2': {'CAS No.': 7782505,
         'ZPE': 0.7737,
         'charges': None,
         'database': 'G2-1',
         'description': 'Cl2 molecule, D*h symm.',
         'enthalpy': 0.0,
         'ionization energy': 11.48,
         'magmoms': None,
         'name': 'Cl_2',
         'positions': [[0.0, 0.0, 1.007541], [0.0, 0.0, -1.007541]],
         'symbols': 'ClCl',
         'thermal correction': 2.1963,
         'vertical ionization energy': 11.49},
 'ClF': {'CAS No.': 7790898,
         'ZPE': 1.1113,
         'charges': None,
         'database': 'G2-1',
         'description': 'ClF molecule, C*v symm, 1-SG.',
         'enthalpy': -13.2,
         'ionization energy': 12.66,
         'magmoms': None,
         'name': 'FCl',
         'positions': [[0.0, 0.0, -1.084794], [0.0, 0.0, 0.574302]],
         'symbols': 'FCl',
         'thermal correction': 2.1273,
         'vertical ionization energy': 12.77},
 'ClO': {'CAS No.': 14989301,
         'ZPE': 1.1923,
         'charges': None,
         'database': 'G2-1',
         'description': 'ClO radical, C*v symm, 2-PI.',
         'enthalpy': 24.2,
         'ionization energy': 10.89,
         'magmoms': [1.0, 0.0],
         'name': 'ClO',
         'positions': [[0.0, 0.0, 0.514172], [0.0, 0.0, -1.092615]],
         'symbols': 'ClO',
         'thermal correction': 2.1172,
         'vertical ionization energy': 11.01},
 'F': {'CAS No.': 14762948,
       'charges': None,
       'database': 'G2-1',
       'description': 'F atom',
       'enthalpy': 18.47,
       'ionization energy': 17.42,
       'magmoms': [1.0],
       'name': 'Fluorine',
       'positions': [[0.0, 0.0, 0.0]],
       'symbols': 'F',
       'thermal correction': 1.05},
 'F2': {'CAS No.': 7782414,
        'ZPE': 1.5179,
        'charges': None,
        'database': 'G2-1',
        'description': 'F2 molecule, D*h symm.',
        'enthalpy': 0.0,
        'ionization energy': 15.7,
        'magmoms': None,
        'name': 'F_2',
        'positions': [[0.0, 0.0, 0.710304], [0.0, 0.0, -0.710304]],
        'symbols': 'FF',
        'thermal correction': 2.0915,
        'vertical ionization energy': 15.7},
 'H': {'CAS No.': 12385136,
       'charges': None,
       'database': 'G2-1',
       'description': 'H atom',
       'enthalpy': 51.63,
       'ionization energy': 13.6,
       'magmoms': [1.0],
       'name': 'Hydrogen',
       'positions': [[0.0, 0.0, 0.0]],
       'symbols': 'H',
       'thermal correction': 1.01},
 'H2CO': {'CAS No.': 50000,
          'ZPE': 16.4502,
          'charges': None,
          'database': 'G2-1',
          'description': 'Formaldehyde (H2C=O), C2v symm.',
          'enthalpy': -26.0,
          'ionization energy': 10.88,
          'magmoms': None,
          'name': 'H_2CO',
          'positions': [[0.0, 0.0, 0.683501],
                        [0.0, 0.0, -0.536614],
                        [0.0, 0.93439, -1.124164],
                        [0.0, -0.93439, -1.124164]],
          'symbols': 'OCHH',
          'thermal correction': 2.3927,
          'vertical ionization energy': 10.88},
 'H2O': {'CAS No.': 7732185,
         'ZPE': 13.2179,
         'charges': None,
         'database': 'G2-1',
         'description': 'Water (H2O), C2v symm.',
         'enthalpy': -57.8,
         'ionization energy': 12.62,
         'magmoms': None,
         'name': 'H_2O',
         'positions': [[0.0, 0.0, 0.119262],
                       [0.0, 0.763239, -0.477047],
                       [0.0, -0.763239, -0.477047]],
         'symbols': 'OHH',
         'thermal correction': 2.372},
 'H2O2': {'CAS No.': 7722841,
          'ZPE': 16.4081,
          'charges': None,
          'database': 'G2-1',
          'description': 'Hydrogen peroxide (HO-OH), C2 symm.',
          'enthalpy': -32.5,
          'ionization energy': 10.58,
          'magmoms': None,
          'name': 'HOOH',
          'positions': [[0.0, 0.734058, -0.05275],
                        [0.0, -0.734058, -0.05275],
                        [0.839547, 0.880752, 0.422001],
                        [-0.839547, -0.880752, 0.422001]],
          'symbols': 'OOHH',
          'thermal correction': 2.623,
          'vertical ionization energy': 11.7},
 'HCN': {'CAS No.': 74908,
         'ZPE': 10.2654,
         'charges': None,
         'database': 'G2-1',
         'description': 'Hydrogen cyanide (HCN), C*v symm.',
         'enthalpy': 31.5,
         'ionization energy': 13.6,
         'magmoms': None,
         'name': 'HCN',
         'positions': [[0.0, 0.0, -0.511747],
                       [0.0, 0.0, 0.664461],
                       [0.0, 0.0, -1.580746]],
         'symbols': 'CNH',
         'thermal correction': 2.1768,
         'vertical ionization energy': 13.61},
 'HCO': {'CAS No.': 2597446,
         'ZPE': 8.029,
         'charges': None,
         'database': 'G2-1',
         'description': 'HCO radical, Bent Cs symm.',
         'enthalpy': 10.0,
         'ionization energy': 8.12,
         'magmoms': [1.0, 0.0, 0.0],
         'name': 'HCO',
         'positions': [[0.06256, 0.593926, 0.0],
                       [0.06256, -0.596914, 0.0],
                       [-0.875835, 1.211755, 0.0]],
         'symbols': 'COH',
         'thermal correction': 2.3864,
         'vertical ionization energy': 9.31},
 'HCl': {'CAS No.': 7647010,
         'ZPE': 4.1673,
         'charges': None,
         'database': 'G2-1',
         'description': 'Hydrogen chloride (HCl), C*v symm.',
         'enthalpy': -22.1,
         'ionization energy': 12.74,
         'magmoms': None,
         'name': 'HCl',
         'positions': [[0.0, 0.0, 0.07111], [0.0, 0.0, -1.208868]],
         'symbols': 'ClH',
         'thermal correction': 2.0739},
 'HF': {'CAS No.': 7664393,
        'ZPE': 5.7994,
        'charges': None,
        'database': 'G2-1',
        'description': 'Hydrogen fluoride (HF), C*v symm.',
        'enthalpy': -65.1,
        'ionization energy': 16.03,
        'magmoms': None,
        'name': 'HF',
        'positions': [[0.0, 0.0, 0.093389], [0.0, 0.0, -0.840502]],
        'symbols': 'FH',
        'thermal correction': 2.0733,
        'vertical ionization energy': 16.12},
 'HOCl': {'CAS No.': 7790923,
          'ZPE': 8.1539,
          'charges': None,
          'database': 'G2-1',
          'description': 'HOCl molecule, Cs symm.',
          'enthalpy': -17.8,
          'ionization energy': 11.12,
          'magmoms': None,
          'name': 'HOCl (hypochlorous acid)',
          'positions': [[0.036702, 1.113517, 0.0],
                        [-0.917548, 1.328879, 0.0],
                        [0.036702, -0.602177, 0.0]],
          'symbols': 'OHCl',
          'thermal correction': 2.4416},
 'Li': {'CAS No.': 7439932,
        'charges': None,
        'database': 'G2-1',
        'description': 'Li atom',
        'enthalpy': 37.69,
        'ionization energy': 5.39,
        'magmoms': [1.0],
        'name': 'Lithium',
        'positions': [[0.0, 0.0, 0.0]],
        'symbols': 'Li',
        'thermal correction': 1.1},
 'Li2': {'CAS No.': 14452596,
         'ZPE': 0.4838,
         'charges': None,
         'database': 'G2-1',
         'description': 'Dilithium (Li2), D*h symm.',
         'enthalpy': 51.6,
         'ionization energy': 5.11,
         'magmoms': None,
         'name': 'Li_2',
         'positions': [[0.0, 0.0, 1.38653], [0.0, 0.0, -1.38653]],
         'symbols': 'LiLi',
         'thermal correction': 2.3086},
 'LiF': {'CAS No.': 7789244,
         'ZPE': 1.4019,
         'charges': None,
         'database': 'G2-1',
         'description': 'Lithium Fluoride (LiF), C*v symm.',
         'enthalpy': -80.1,
         'ionization energy': 11.3,
         'magmoms': None,
         'name': 'LiF',
         'positions': [[0.0, 0.0, -1.174965], [0.0, 0.0, 0.391655]],
         'symbols': 'LiF',
         'thermal correction': 2.099},
 'LiH': {'CAS No.': 7580678,
         'ZPE': 2.0149,
         'charges': None,
         'database': 'G2-1',
         'description': 'Lithium hydride (LiH), C*v symm.',
         'enthalpy': 33.3,
         'ionization energy': 7.9,
         'magmoms': None,
         'name': 'LiH',
         'positions': [[0.0, 0.0, 0.41], [0.0, 0.0, -1.23]],
         'symbols': 'LiH',
         'thermal correction': 2.0783},
 'N': {'CAS No.': 17778880,
       'charges': None,
       'database': 'G2-1',
       'description': 'N atom',
       'enthalpy': 112.53,
       'ionization energy': 14.53,
       'magmoms': [3.0],
       'name': 'Nitrogen',
       'positions': [[0.0, 0.0, 0.0]],
       'symbols': 'N',
       'thermal correction': 1.04},
 'N2': {'CAS No.': 7727379,
        'ZPE': 3.4243,
        'charges': None,
        'database': 'G2-1',
        'description': 'N2 molecule, D*h symm.',
        'enthalpy': 0.0,
        'ionization energy': 15.58,
        'magmoms': None,
        'name': 'N_2',
        'positions': [[0.0, 0.0, 0.56499], [0.0, 0.0, -0.56499]],
        'symbols': 'NN',
        'thermal correction': 2.0733,
        'vertical ionization energy': 15.58},
 'N2H4': {'CAS No.': 302012,
          'ZPE': 32.9706,
          'charges': None,
          'database': 'G2-1',
          'description': 'Hydrazine (H2N-NH2), C2 symm.',
          'enthalpy': 22.8,
          'ionization energy': 8.1,
          'magmoms': None,
          'name': 'H_2NNH_2',
          'positions': [[0.0, 0.718959, -0.077687],
                        [0.0, -0.718959, -0.077687],
                        [0.211082, 1.092752, 0.847887],
                        [-0.948214, 1.005026, -0.304078],
                        [-0.211082, -1.092752, 0.847887],
                        [0.948214, -1.005026, -0.304078]],
          'symbols': 'NNHHHH',
          'thermal correction': 2.6531,
          'vertical ionization energy': 8.98},
 'NH': {'CAS No.': 13774920,
        'ZPE': 4.5739,
        'charges': None,
        'database': 'G2-1',
        'description': 'NH, triplet, C*v symm.',
        'enthalpy': 85.2,
        'ionization energy': 13.1,
        'magmoms': [2.0, 0.0],
        'name': 'NH',
        'positions': [[0.0, 0.0, 0.129929], [0.0, 0.0, -0.909501]],
        'symbols': 'NH',
        'thermal correction': 2.0739,
        'vertical ionization energy': 13.49},
 'NH2': {'CAS No.': 13770406,
         'ZPE': 11.742,
         'charges': None,
         'database': 'G2-1',
         'description': 'NH2 radical, C2v symm, 2-B1.',
         'enthalpy': 45.1,
         'ionization energy': 10.78,
         'magmoms': [1.0, 0.0, 0.0],
         'name': 'NH_2',
         'positions': [[0.0, 0.0, 0.14169],
                       [0.0, 0.806442, -0.495913],
                       [0.0, -0.806442, -0.495913]],
         'symbols': 'NHH',
         'thermal correction': 2.3726,
         'vertical ionization energy': 12.0},
 'NH3': {'CAS No.': 7664417,
         'ZPE': 21.2462,
         'charges': None,
         'database': 'G2-1',
         'description': 'Ammonia (NH3), C3v symm.',
         'enthalpy': -11.0,
         'ionization energy': 10.07,
         'magmoms': None,
         'name': 'NH_3',
         'positions': [[0.0, 0.0, 0.116489],
                       [0.0, 0.939731, -0.271808],
                       [0.813831, -0.469865, -0.271808],
                       [-0.813831, -0.469865, -0.271808]],
         'symbols': 'NHHH',
         'thermal correction': 2.3896,
         'vertical ionization energy': 10.82},
 'NO': {'CAS No.': 10102439,
        'ZPE': 2.7974,
        'charges': None,
        'database': 'G2-1',
        'description': 'NO radical, C*v symm, 2-Pi.',
        'enthalpy': 21.6,
        'ionization energy': 9.26,
        'magmoms': [0.6, 0.4],
        'name': 'NO',
        'positions': [[0.0, 0.0, -0.609442], [0.0, 0.0, 0.533261]],
        'symbols': 'NO',
        'thermal correction': 2.0745,
        'vertical ionization energy': 9.26},
 'Na': {'CAS No.': 7440235,
        'charges': None,
        'database': 'G2-1',
        'description': 'Na atom',
        'enthalpy': 25.69,
        'ionization energy': 5.14,
        'magmoms': [1.0],
        'name': 'Sodium',
        'positions': [[0.0, 0.0, 0.0]],
        'symbols': 'Na',
        'thermal correction': 1.54},
 'Na2': {'CAS No.': 25681792,
         'ZPE': 0.2246,
         'charges': None,
         'database': 'G2-1',
         'description': 'Disodium (Na2), D*h symm.',
         'enthalpy': 34.0,
         'ionization energy': 4.89,
         'magmoms': None,
         'name': 'Na_2',
         'positions': [[0.0, 0.0, 1.576262], [0.0, 0.0, -1.576262]],
         'symbols': 'NaNa',
         'thermal correction': 2.4699},
 'NaCl': {'CAS No.': 7647145,
          'ZPE': 0.5152,
          'charges': None,
          'database': 'G2-1',
          'description': 'Sodium Chloride (NaCl), C*v symm.',
          'enthalpy': -43.6,
          'ionization energy': 9.2,
          'magmoms': None,
          'name': 'NaCl',
          'positions': [[0.0, 0.0, -1.45166], [0.0, 0.0, 0.93931]],
          'symbols': 'NaCl',
          'thermal correction': 2.2935,
          'vertical ionization energy': 9.8},
 'O': {'CAS No.': 17778802,
       'charges': None,
       'database': 'G2-1',
       'description': 'O atom',
       'enthalpy': 58.99,
       'ionization energy': 13.62,
       'magmoms': [2.0],
       'name': 'Oxygen',
       'positions': [[0.0, 0.0, 0.0]],
       'symbols': 'O',
       'thermal correction': 1.04},
 'O2': {'CAS No.': 7782447,
        'ZPE': 2.3444,
        'charges': None,
        'database': 'G2-1',
        'description': 'O2 molecule, D*h symm, Triplet.',
        'enthalpy': 0.0,
        'ionization energy': 12.07,
        'magmoms': [1.0, 1.0],
        'name': 'O_2',
        'positions': [[0.0, 0.0, 0.622978], [0.0, 0.0, -0.622978]],
        'symbols': 'OO',
        'thermal correction': 2.0752,
        'vertical ionization energy': 12.3},
 'OH': {'CAS No.': 3352576,
        'ZPE': 5.2039,
        'charges': None,
        'database': 'G2-1',
        'description': 'OH radical, C*v symm.',
        'enthalpy': 9.4,
        'ionization energy': 13.02,
        'magmoms': [0.5, 0.5],
        'name': 'OH',
        'positions': [[0.0, 0.0, 0.108786], [0.0, 0.0, -0.870284]],
        'symbols': 'OH',
        'thermal correction': 2.0739},
 'P': {'CAS No.': 7723140,
       'charges': None,
       'database': 'G2-1',
       'description': 'P atom',
       'enthalpy': 75.42,
       'ionization energy': 10.49,
       'magmoms': [3.0],
       'name': 'Phosphorus',
       'positions': [[0.0, 0.0, 0.0]],
       'symbols': 'P',
       'thermal correction': 1.28},
 'P2': {'CAS No.': 12185090,
        'ZPE': 1.1358,
        'charges': None,
        'database': 'G2-1',
        'description': 'P2 molecule, D*h symm.',
        'enthalpy': 34.3,
        'ionization energy': 10.53,
        'magmoms': None,
        'name': 'P_2',
        'positions': [[0.0, 0.0, 0.966144], [0.0, 0.0, -0.966144]],
        'symbols': 'PP',
        'thermal correction': 2.1235,
        'vertical ionization energy': 10.62},
 'PH2': {'CAS No.': 13765430,
         'ZPE': 8.2725,
         'charges': None,
         'database': 'G2-1',
         'description': 'PH2 radical, C2v symm.',
         'enthalpy': 33.1,
         'ionization energy': 9.82,
         'magmoms': [1.0, 0.0, 0.0],
         'name': 'PH_2 (Phosphino radical)',
         'positions': [[0.0, 0.0, 0.115396],
                       [0.0, 1.025642, -0.865468],
                       [0.0, -1.025642, -0.865468]],
         'symbols': 'PHH',
         'thermal correction': 2.3845},
 'PH3': {'CAS No.': 7803512,
         'ZPE': 14.7885,
         'charges': None,
         'database': 'G2-1',
         'description': 'Phosphine (PH3), C3v symm.',
         'enthalpy': 1.3,
         'ionization energy': 9.87,
         'magmoms': None,
         'name': 'PH_3',
         'positions': [[0.0, 0.0, 0.124619],
                       [0.0, 1.200647, -0.623095],
                       [1.039791, -0.600323, -0.623095],
                       [-1.039791, -0.600323, -0.623095]],
         'symbols': 'PHHH',
         'thermal correction': 2.4203,
         'vertical ionization energy': 10.95},
 'S': {'CAS No.': 7704349,
       'charges': None,
       'database': 'G2-1',
       'description': 'S atom',
       'enthalpy': 65.66,
       'ionization energy': 10.36,
       'magmoms': [2.0],
       'name': 'Sulfur',
       'positions': [[0.0, 0.0, 0.0]],
       'symbols': 'S',
       'thermal correction': 1.05},
 'S2': {'CAS No.': 23550450,
        'ZPE': 1.0078,
        'charges': None,
        'database': 'G2-1',
        'description': 'S2 molecule, D*h symm, triplet.',
        'enthalpy': 30.7,
        'ionization energy': 9.36,
        'magmoms': [1.0, 1.0],
        'name': 'S_2',
        'positions': [[0.0, 0.0, 0.960113], [0.0, 0.0, -0.960113]],
        'symbols': 'SS',
        'thermal correction': 2.1436,
        'vertical ionization energy': 9.55},
 'SH2': {'CAS No.': 7783064,
         'ZPE': 9.3129,
         'charges': None,
         'database': 'G2-1',
         'description': 'Hydrogen sulfide (H2S), C2v symm.',
         'enthalpy': -4.9,
         'ionization energy': 10.46,
         'magmoms': None,
         'name': 'SH_2',
         'positions': [[0.0, 0.0, 0.102135],
                       [0.0, 0.974269, -0.817083],
                       [0.0, -0.974269, -0.817083]],
         'symbols': 'SHH',
         'thermal correction': 2.3808,
         'vertical ionization energy': 10.5},
 'SO': {'CAS No.': 13827322,
        'ZPE': 1.6158,
        'charges': None,
        'database': 'G2-1',
        'description': 'Sulfur monoxide (SO), C*v symm, triplet.',
        'enthalpy': 1.2,
        'ionization energy': 11.29,
        'magmoms': [1.0, 1.0],
        'name': 'SO',
        'positions': [[0.0, 0.0, -1.015992], [0.0, 0.0, 0.507996]],
        'symbols': 'OS',
        'thermal correction': 2.0877},
 'SO2': {'CAS No.': 7446095,
         'ZPE': 4.3242,
         'charges': None,
         'database': 'G2-1',
         'description': 'Sulfur dioxide (SO2), C2v symm.',
         'enthalpy': -71.0,
         'ionization energy': 12.35,
         'magmoms': None,
         'name': 'SO_2',
         'positions': [[0.0, 0.0, 0.370268],
                       [0.0, 1.277617, -0.370268],
                       [0.0, -1.277617, -0.370268]],
         'symbols': 'SOO',
         'thermal correction': 2.5245,
         'vertical ionization energy': 12.5},
 'Si': {'CAS No.': 7440213,
        'charges': None,
        'database': 'G2-1',
        'description': 'Si atom',
        'enthalpy': 106.6,
        'ionization energy': 8.15,
        'magmoms': [2.0],
        'name': 'Silicon',
        'positions': [[0.0, 0.0, 0.0]],
        'symbols': 'Si',
        'thermal correction': 0.76},
 'Si2': {'CAS No.': 12597352,
         'ZPE': 0.7028,
         'charges': None,
         'database': 'G2-1',
         'description': 'Si2 molecule, D*h symm, Triplet (3-Sigma-G-).',
         'enthalpy': 139.9,
         'ionization energy': 7.9,
         'magmoms': [1.0, 1.0],
         'name': 'Si_2 (Silicon diatomic)',
         'positions': [[0.0, 0.0, 1.130054], [0.0, 0.0, -1.130054]],
         'symbols': 'SiSi',
         'thermal correction': 2.2182},
 'Si2H6': {'CAS No.': 1590870,
           'ZPE': 30.2265,
           'charges': None,
           'database': 'G2-1',
           'description': 'Disilane (H3Si-SiH3), D3d symm.',
           'enthalpy': 19.1,
           'ionization energy': 9.74,
           'magmoms': None,
           'name': 'Si_2H_6',
           'positions': [[0.0, 0.0, 1.167683],
                         [0.0, 0.0, -1.167683],
                         [0.0, 1.393286, 1.68602],
                         [-1.206621, -0.696643, 1.68602],
                         [1.206621, -0.696643, 1.68602],
                         [0.0, -1.393286, -1.68602],
                         [-1.206621, 0.696643, -1.68602],
                         [1.206621, 0.696643, -1.68602]],
           'symbols': 'SiSiHHHHHH',
           'thermal correction': 3.7927,
           'vertical ionization energy': 10.53},
 'SiH2_s1A1d': {'CAS No.': 13825906,
                'ZPE': 7.1875,
                'charges': None,
                'database': 'G2-1',
                'description': 'Singlet silylene (SiH2), C2v symm, 1-A1.',
                'enthalpy': 65.2,
                'ionization energy': 8.92,
                'magmoms': None,
                'name': 'SiH_2 (^1A_1)(silicon dihydride)',
                'positions': [[0.0, 0.0, 0.131272],
                              [0.0, 1.096938, -0.918905],
                              [0.0, -1.096938, -0.918905]],
                'symbols': 'SiHH',
                'thermal correction': 2.3927},
 'SiH2_s3B1d': {'CAS No.': 13825906,
                'ZPE': 7.4203,
                'charges': None,
                'database': 'G2-1',
                'description': 'Triplet silylene (SiH2), C2v symm, 3-B1.',
                'enthalpy': 86.2,
                'magmoms': [2.0, 0.0, 0.0],
                'name': 'SiH_2 (^3B_1)(silicon dihydride)',
                'positions': [[0.0, 0.0, 0.094869],
                              [0.0, 1.271862, -0.664083],
                              [0.0, -1.271862, -0.664083]],
                'symbols': 'SiHH',
                'thermal correction': 2.4078},
 'SiH3': {'CAS No.': 13765441,
          'ZPE': 13.0898,
          'charges': None,
          'database': 'G2-1',
          'description': 'Silyl radical (SiH3), C3v symm.',
          'enthalpy': 47.9,
          'ionization energy': 8.14,
          'magmoms': [1.0, 0.0, 0.0, 0.0],
          'name': 'SiH_3',
          'positions': [[0.0, 0.0, 0.079299],
                        [0.0, 1.41328, -0.370061],
                        [1.223937, -0.70664, -0.370061],
                        [-1.223937, -0.70664, -0.370061]],
          'symbols': 'SiHHH',
          'thermal correction': 2.4912,
          'vertical ionization energy': 8.74},
 'SiH4': {'CAS No.': 7803625,
          'ZPE': 19.2664,
          'charges': None,
          'database': 'G2-1',
          'description': 'Silane (SiH4), Td symm.',
          'enthalpy': 8.2,
          'ionization energy': 11.0,
          'magmoms': None,
          'name': 'SiH_4',
          'positions': [[0.0, 0.0, 0.0],
                        [0.856135, 0.856135, 0.856135],
                        [-0.856135, -0.856135, 0.856135],
                        [-0.856135, 0.856135, -0.856135],
                        [0.856135, -0.856135, -0.856135]],
          'symbols': 'SiHHHH',
          'thermal correction': 2.5232,
          'vertical ionization energy': 12.3},
 'SiO': {'CAS No.': 10097286,
         'ZPE': 1.7859,
         'charges': None,
         'database': 'G2-1',
         'description': 'Silicon monoxide (SiO), C*v symm.',
         'enthalpy': -24.6,
         'ionization energy': 11.49,
         'magmoms': None,
         'name': 'SiO',
         'positions': [[0.0, 0.0, 0.560846], [0.0, 0.0, -0.98148]],
         'symbols': 'SiO',
         'thermal correction': 2.0821}}

def get_ionization_energy(name, vertical=True):
    """Return the experimental ionization energy from the database.

    If vertical is True, the vertical ionization energy is returned if
    available.
    """
    if name not in data:
        raise KeyError('System %s not in database.' % name)
    elif 'ionization energy' not in data[name]:
        raise KeyError('No data on ionization energy for system %s.' % name)
    else:
        if vertical and 'vertical ionization energy' in data[name]:
            return data[name]['vertical ionization energy']
        else:
            return data[name]['ionization energy']


def get_atomization_energy(name):
    """Determine extrapolated experimental atomization energy from the database.

    The atomization energy is extrapolated from experimental heats of
    formation at room temperature, using calculated zero-point energies
    and thermal corrections.

    The atomization energy is returned in kcal/mol = 43.36 meV:

    >>> from ase.units import *; print kcal / mol
    0.0433641146392

    """
    d = data[name]
    e = d['enthalpy']
    z = d['ZPE']
    dh = d['thermal correction']
    ae = -e + z + dh
    for a in string2symbols(d['symbols']):
        h = data[a]['enthalpy']
        dh = data[a]['thermal correction']
        ae += h - dh
    return ae



# -*- coding: utf-8 -*-
"""Vibrational modes."""

import os
import os.path as op
import pickle
import sys
from math import sin, pi, sqrt, log

import numpy as np

import ase.units as units
from ase.io.trajectory import Trajectory
from ase.parallel import world, paropen

from ase.utils import opencew, pickleload, basestring
from ase.calculators.singlepoint import SinglePointCalculator


class Vibrations:
    """Class for calculating vibrational modes using finite difference.

    The vibrational modes are calculated from a finite difference
    approximation of the Hessian matrix.

    The *summary()*, *get_energies()* and *get_frequencies()* methods all take
    an optional *method* keyword.  Use method='Frederiksen' to use the method
    described in:

      T. Frederiksen, M. Paulsson, M. Brandbyge, A. P. Jauho:
      "Inelastic transport theory from first-principles: methodology and
      applications for nanoscale devices", Phys. Rev. B 75, 205413 (2007)

    atoms: Atoms object
        The atoms to work on.
    indices: list of int
        List of indices of atoms to vibrate.  Default behavior is
        to vibrate all atoms.
    name: str
        Name to use for files.
    delta: float
        Magnitude of displacements.
    nfree: int
        Number of displacements per atom and cartesian coordinate, 2 and 4 are
        supported. Default is 2 which will displace each atom +delta and
        -delta for each cartesian coordinate.

    Example:

    >>> from ase import Atoms
    >>> from ase.calculators.emt import EMT
    >>> from ase.optimize import BFGS
    >>> from ase.vibrations import Vibrations
    >>> n2 = Atoms('N2', [(0, 0, 0), (0, 0, 1.1)],
    ...            calculator=EMT())
    >>> BFGS(n2).run(fmax=0.01)
    BFGS:   0  16:01:21        0.440339       3.2518
    BFGS:   1  16:01:21        0.271928       0.8211
    BFGS:   2  16:01:21        0.263278       0.1994
    BFGS:   3  16:01:21        0.262777       0.0088
    >>> vib = Vibrations(n2)
    >>> vib.run()
    Writing vib.eq.pckl
    Writing vib.0x-.pckl
    Writing vib.0x+.pckl
    Writing vib.0y-.pckl
    Writing vib.0y+.pckl
    Writing vib.0z-.pckl
    Writing vib.0z+.pckl
    Writing vib.1x-.pckl
    Writing vib.1x+.pckl
    Writing vib.1y-.pckl
    Writing vib.1y+.pckl
    Writing vib.1z-.pckl
    Writing vib.1z+.pckl
    >>> vib.summary()
    ---------------------
    #    meV     cm^-1
    ---------------------
    0    0.0       0.0
    1    0.0       0.0
    2    0.0       0.0
    3    2.5      20.4
    4    2.5      20.4
    5  152.6    1230.8
    ---------------------
    Zero-point energy: 0.079 eV
    >>> vib.write_mode(-1)  # write last mode to trajectory file

    """

    def __init__(self, atoms, indices=None, name='vib', delta=0.01, nfree=2):
        assert nfree in [2, 4]
        self.atoms = atoms
        self.calc = atoms.get_calculator()
        if indices is None:
            indices = range(len(atoms))
        self.indices = np.asarray(indices)
        self.name = name
        self.delta = delta
        self.nfree = nfree
        self.H = None
        self.ir = None
        self.ram = None

    def run(self):
        """Run the vibration calculations.

        This will calculate the forces for 6 displacements per atom +/-x,
        +/-y, +/-z. Only those calculations that are not already done will be
        started. Be aware that an interrupted calculation may produce an empty
        file (ending with .pckl), which must be deleted before restarting the
        job. Otherwise the forces will not be calculated for that
        displacement.

        Note that the calculations for the different displacements can be done
        simultaneously by several independent processes. This feature relies
        on the existence of files and the subsequent creation of the file in
        case it is not found.

        If the program you want to use does not have a calculator in ASE, use
        ``iterdisplace`` to get all displaced structures and calculate the forces
        on your own.
        """

        if op.isfile(self.name + '.all.pckl'):
            raise RuntimeError(
                'Cannot run calculation. ' +
                self.name + '.all.pckl must be removed or split in order ' +
                'to have only one sort of data structure at a time.')
        for dispName, atoms in self.iterdisplace(inplace=True):
            filename = dispName + '.pckl'
            fd = opencew(filename)
            if fd is not None:
                self.calculate(atoms, filename, fd)

    def iterdisplace(self, inplace=False):
        """Yield name and atoms object for initial and displaced structures.

        Use this to export the structures for each single-point calculation
        to an external program instead of using ``run()``. Then save the
        calculated gradients to <name>.pckl and continue using this instance.
        """
        atoms = self.atoms if inplace else self.atoms.copy()
        yield self.name + '.eq', atoms
        for dispName, a, i, disp in self.displacements():
            if not inplace:
                atoms = self.atoms.copy()
            pos0 = atoms.positions[a, i]
            atoms.positions[a, i] += disp
            yield dispName, atoms
            if inplace:
                atoms.positions[a, i] = pos0

    def iterimages(self):
        """Yield initial and displaced structures."""
        for name, atoms in self.iterdisplace():
            yield atoms

    def displacements(self):
        for a in self.indices:
            for i in range(3):
                for sign in [-1, 1]:
                    for ndis in range(1, self.nfree // 2 + 1):
                        dispName = ('%s.%d%s%s' %
                                    (self.name, a, 'xyz'[i],
                                     ndis * ' +-'[sign]))
                        disp = ndis * sign * self.delta
                        yield dispName, a, i, disp

    def calculate(self, atoms, filename, fd):
        forces = self.calc.get_forces(atoms)
        if self.ir:
            dipole = self.calc.get_dipole_moment(atoms)
        if self.ram:
            freq, noninPol, pol = self.get_polarizability()
        if world.rank == 0:
            if self.ir and self.ram:
                pickle.dump([forces, dipole, freq, noninPol, pol], fd, protocol=2)
                sys.stdout.write(
                    'Writing %s, dipole moment = (%.6f %.6f %.6f)\n' %
                    (filename, dipole[0], dipole[1], dipole[2]))
            elif self.ir and not self.ram:
                pickle.dump([forces, dipole], fd, protocol=2)
                sys.stdout.write(
                    'Writing %s, dipole moment = (%.6f %.6f %.6f)\n' %
                    (filename, dipole[0], dipole[1], dipole[2]))
            else:
                pickle.dump(forces, fd, protocol=2)
                sys.stdout.write('Writing %s\n' % filename)
            fd.close()
        sys.stdout.flush()

    def clean(self, empty_files=False, combined=True):
        """Remove pickle-files.

        Use empty_files=True to remove only empty files and
        combined=False to not remove the combined file.

        """

        if world.rank != 0:
            return 0

        n = 0
        filenames = [self.name + '.eq.pckl']
        if combined:
            filenames.append(self.name + '.all.pckl')
        for dispName, a, i, disp in self.displacements():
            filename = dispName + '.pckl'
            filenames.append(filename)

        for name in filenames:
            if op.isfile(name):
                if not empty_files or op.getsize(name) == 0:
                    os.remove(name)
                    n += 1
        return n

    def combine(self):
        """Combine pickle-files to one file ending with '.all.pckl'.

        The other pickle-files will be removed in order to have only one sort
        of data structure at a time.

        """
        if world.rank != 0:
            return 0
        filenames = [self.name + '.eq.pckl']
        for dispName, a, i, disp in self.displacements():
            filename = dispName + '.pckl'
            filenames.append(filename)
        combined_data = {}
        for name in filenames:
            if not op.isfile(name) or op.getsize(name) == 0:
                raise RuntimeError('Calculation is not complete. ' +
                                   name + ' is missing or empty.')
            with open(name, 'rb') as fl:
                f = pickleload(fl)
            combined_data.update({op.basename(name): f})
        filename = self.name + '.all.pckl'
        fd = opencew(filename)
        if fd is None:
            raise RuntimeError(
                'Cannot write file ' + filename +
                '. Remove old file if it exists.')
        else:
            pickle.dump(combined_data, fd, protocol=2)
            fd.close()
        return self.clean(combined=False)

    def split(self):
        """Split combined pickle-file.

        The combined pickle-file will be removed in order to have only one
        sort of data structure at a time.

        """
        if world.rank != 0:
            return 0
        combined_name = self.name + '.all.pckl'
        if not op.isfile(combined_name):
            raise RuntimeError('Cannot find combined file: ' +
                               combined_name + '.')
        with open(combined_name, 'rb') as fl:
            combined_data = pickleload(fl)
        filenames = [self.name + '.eq.pckl']
        for dispName, a, i, disp in self.displacements():
            filename = dispName + '.pckl'
            filenames.append(filename)
            if op.isfile(filename):
                raise RuntimeError(
                    'Cannot split. File ' + filename + 'already exists.')
        for name in filenames:
            fd = opencew(name)
            try:
                pickle.dump(combined_data[op.basename(name)], fd, protocol=2)
            except KeyError:
                pickle.dump(combined_data[name], fd, protocol=2)  # Old version
            fd.close()
        os.remove(combined_name)
        return 1  # One file removed

    def read(self, method='standard', direction='central'):
        self.method = method.lower()
        self.direction = direction.lower()
        assert self.method in ['standard', 'frederiksen']
        assert self.direction in ['central', 'forward', 'backward']

        def load(fname, combined_data=None):
            if combined_data is None:
                with open(fname, 'rb') as fl:
                    f = pickleload(fl)
            else:
                try:
                    f = combined_data[op.basename(fname)]
                except KeyError:
                    f = combined_data[fname]  # Old version
            if not hasattr(f, 'shape') and not hasattr(f, 'keys'):
                # output from InfraRed
                return f[0]
            return f

        n = 3 * len(self.indices)
        H = np.empty((n, n))
        r = 0
        if op.isfile(self.name + '.all.pckl'):
            # Open the combined pickle-file
            combined_data = load(self.name + '.all.pckl')
        else:
            combined_data = None
        if direction != 'central':
            feq = load(self.name + '.eq.pckl', combined_data)
        for a in self.indices:
            for i in 'xyz':
                name = '%s.%d%s' % (self.name, a, i)
                fminus = load(name + '-.pckl', combined_data)
                fplus = load(name + '+.pckl', combined_data)
                if self.method == 'frederiksen':
                    fminus[a] -= fminus.sum(0)
                    fplus[a] -= fplus.sum(0)
                if self.nfree == 4:
                    fminusminus = load(name + '--.pckl', combined_data)
                    fplusplus = load(name + '++.pckl', combined_data)
                    if self.method == 'frederiksen':
                        fminusminus[a] -= fminusminus.sum(0)
                        fplusplus[a] -= fplusplus.sum(0)
                if self.direction == 'central':
                    if self.nfree == 2:
                        H[r] = .5 * (fminus - fplus)[self.indices].ravel()
                    else:
                        H[r] = H[r] = (-fminusminus +
                                       8 * fminus -
                                       8 * fplus +
                                       fplusplus)[self.indices].ravel() / 12.0
                elif self.direction == 'forward':
                    H[r] = (feq - fplus)[self.indices].ravel()
                else:
                    assert self.direction == 'backward'
                    H[r] = (fminus - feq)[self.indices].ravel()
                H[r] /= 2 * self.delta
                r += 1
        H += H.copy().T
        self.H = H
        m = self.atoms.get_masses()
        if 0 in [m[index] for index in self.indices]:
            raise RuntimeError('Zero mass encountered in one or more of '
                               'the vibrated atoms. Use Atoms.set_masses()'
                               ' to set all masses to non-zero values.')

        self.im = np.repeat(m[self.indices]**-0.5, 3)
        omega2, modes = np.linalg.eigh(self.im[:, None] * H * self.im)
        self.modes = modes.T.copy()

        # Conversion factor:
        s = units._hbar * 1e10 / sqrt(units._e * units._amu)
        self.hnu = s * omega2.astype(complex)**0.5

    def get_energies(self, method='standard', direction='central', **kw):
        """Get vibration energies in eV."""

        if (self.H is None or method.lower() != self.method or
            direction.lower() != self.direction):
            self.read(method, direction, **kw)
        return self.hnu

    def get_frequencies(self, method='standard', direction='central'):
        """Get vibration frequencies in cm^-1."""

        s = 1. / units.invcm
        return s * self.get_energies(method, direction)

    def summary(self, method='standard', direction='central', freq=None,
                log=sys.stdout):
        """Print a summary of the vibrational frequencies.

        Parameters:

        method : string
            Can be 'standard'(default) or 'Frederiksen'.
        direction: string
            Direction for finite differences. Can be one of 'central'
            (default), 'forward', 'backward'.
        freq : numpy array
            Optional. Can be used to create a summary on a set of known
            frequencies.
        log : if specified, write output to a different location than
            stdout. Can be an object with a write() method or the name of a
            file to create.
        """

        if isinstance(log, basestring):
            log = paropen(log, 'a')
        write = log.write

        s = 0.01 * units._e / units._c / units._hplanck
        if freq is not None:
            hnu = freq / s
        else:
            hnu = self.get_energies(method, direction)
        write('---------------------\n')
        write('  #    meV     cm^-1\n')
        write('---------------------\n')
        for n, e in enumerate(hnu):
            if e.imag != 0:
                c = 'i'
                e = e.imag
            else:
                c = ' '
                e = e.real
            write('%3d %6.1f%s  %7.1f%s\n' % (n, 1000 * e, c, s * e, c))
        write('---------------------\n')
        write('Zero-point energy: %.3f eV\n' %
              self.get_zero_point_energy(freq=freq))

    def get_zero_point_energy(self, freq=None):
        if freq is None:
            return 0.5 * self.hnu.real.sum()
        else:
            s = 0.01 * units._e / units._c / units._hplanck
            return 0.5 * freq.real.sum() / s

    def get_mode(self, n):
        """Get mode number ."""
        mode = np.zeros((len(self.atoms), 3))
        mode[self.indices] = (self.modes[n] * self.im).reshape((-1, 3))
        return mode

    def write_mode(self, n=None, kT=units.kB * 300, nimages=30):
        """Write mode number n to trajectory file. If n is not specified,
        writes all non-zero modes."""
        if n is None:
            for index, energy in enumerate(self.get_energies()):
                if abs(energy) > 1e-5:
                    self.write_mode(n=index, kT=kT, nimages=nimages)
            return
        mode = self.get_mode(n) * sqrt(kT / abs(self.hnu[n]))
        p = self.atoms.positions.copy()
        n %= 3 * len(self.indices)
        traj = Trajectory('%s.%d.traj' % (self.name, n), 'w')
        calc = self.atoms.get_calculator()
        self.atoms.set_calculator()
        for x in np.linspace(0, 2 * pi, nimages, endpoint=False):
            self.atoms.set_positions(p + sin(x) * mode)
            traj.write(self.atoms)
        self.atoms.set_positions(p)
        self.atoms.set_calculator(calc)
        traj.close()

    def show_as_force(self, n, scale=0.2):
        mode = self.get_mode(n) * len(self.hnu) * scale
        calc = SinglePointCalculator(self.atoms, forces=mode)
        self.atoms.set_calculator(calc)
        self.atoms.edit()

    def write_jmol(self):
        """Writes file for viewing of the modes with jmol."""

        fd = open(self.name + '.xyz', 'w')
        symbols = self.atoms.get_chemical_symbols()
        f = self.get_frequencies()
        for n in range(3 * len(self.indices)):
            fd.write('%6d\n' % len(self.atoms))
            if f[n].imag != 0:
                c = 'i'
                f[n] = f[n].imag
            else:
                c = ' '
            fd.write('Mode #%d, f = %.1f%s cm^-1' % (n, f[n], c))
            if self.ir:
                fd.write(', I = %.4f (D/Å)^2 amu^-1.\n' % self.intensities[n])
            else:
                fd.write('.\n')
            mode = self.get_mode(n)
            for i, pos in enumerate(self.atoms.positions):
                fd.write('%2s %12.5f %12.5f %12.5f %12.5f %12.5f %12.5f \n' %
                         (symbols[i], pos[0], pos[1], pos[2],
                          mode[i, 0], mode[i, 1], mode[i, 2]))
        fd.close()

    def fold(self, frequencies, intensities,
             start=800.0, end=4000.0, npts=None, width=4.0,
             type='Gaussian', normalize=False):
        """Fold frequencies and intensities within the given range
        and folding method (Gaussian/Lorentzian).
        The energy unit is cm^-1.
        normalize=True ensures the integral over the peaks to give the
        intensity.
        """

        lctype = type.lower()
        assert lctype in ['gaussian', 'lorentzian']
        if not npts:
            npts = int((end - start) / width * 10 + 1)
        prefactor = 1
        if lctype == 'lorentzian':
            intensities = intensities * width * pi / 2.
            if normalize:
                prefactor = 2. / width / pi
        else:
            sigma = width / 2. / sqrt(2. * log(2.))
            if normalize:
                prefactor = 1. / sigma / sqrt(2 * pi)

        # Make array with spectrum data
        spectrum = np.empty(npts)
        energies = np.linspace(start, end, npts)
        for i, energy in enumerate(energies):
            energies[i] = energy
            if lctype == 'lorentzian':
                spectrum[i] = (intensities * 0.5 * width / pi /
                               ((frequencies - energy)**2 +
                                0.25 * width**2)).sum()
            else:
                spectrum[i] = (intensities *
                               np.exp(-(frequencies - energy)**2 /
                                      2. / sigma**2)).sum()
        return [energies, prefactor * spectrum]

    def write_dos(self, out='vib-dos.dat', start=800, end=4000,
                  npts=None, width=10,
                  type='Gaussian', method='standard', direction='central'):
        """Write out the vibrational density of states to file.

        First column is the wavenumber in cm^-1, the second column the
        folded vibrational density of states.
        Start and end points, and width of the Gaussian/Lorentzian
        should be given in cm^-1."""
        frequencies = self.get_frequencies(method, direction).real
        intensities = np.ones(len(frequencies))
        energies, spectrum = self.fold(frequencies, intensities,
                                       start, end, npts, width, type)

        # Write out spectrum in file.
        outdata = np.empty([len(energies), 2])
        outdata.T[0] = energies
        outdata.T[1] = spectrum
        fd = open(out, 'w')
        fd.write('# %s folded, width=%g cm^-1\n' % (type.title(), width))
        fd.write('# [cm^-1] arbitrary\n')
        for row in outdata:
            fd.write('%.3f  %15.5e\n' %
                     (row[0], row[1]))
        fd.close()



from ase.gui.i18n import _

import ase.gui.ui as ui
from ase.utils import rotate, irotate


class Rotate:
    update = True

    def __init__(self, gui):
        self.gui = gui
        win = ui.Window(_('Rotate'))
        win.add(_('Rotation angles:'))
        self.rotate = [ui.SpinBox(42.0, -360, 360, 1, self.change)
                       for i in '123']
        win.add(self.rotate)
        win.add(ui.Button(_('Update'), self.update_angles))
        win.add(_('Note:\nYou can rotate freely\n'
                  'with the mouse, by holding\n'
                  'down mouse button 2.'))
        self.update_angles()

    def change(self):
        x, y, z = [float(a.value) for a in self.rotate]
        self.gui.axes = rotate('%fx,%fy,%fz' % (x, y, z))
        self.gui.set_frame()

    def update_angles(self):
        angles = irotate(self.gui.axes)
        for r, a in zip(self.rotate, angles):
            r.value = a



from ase import Atoms
from ase.db import connect
import ase.db.app as app

c = connect('test.db', append=False)
x = [0, 1, 2]
t1 = [1, 2, 0]
t2 = [[2, 3], [1, 1], [1, 0]]

atoms = Atoms('H2O')
atoms.center(vacuum=5)
atoms.set_pbc(True)

c.write(atoms,
        foo=42.0,
        bar='abc',
        data={'x': x,
              't1': t1,
              't2': t2})
c.metadata = {'title': 'Test title'}
c.python = {'key_descriptions': {'foo': ('FOO', 'FOO ...', '`m_e`')},
            'default_columns': ['foo', 'formula', 'bar']}
app.databases['default'] = c
app.initialize_databases()
app.app.testing = True
c = app.app.test_client()
page = c.get('/').data.decode()
assert 'Test title' in page
assert 'FOO' in page
c.get('/default/row/1')
c.get('/default/json/1').data
c.get('/default/sqlite/1').data
c.get('/default/sqlite?x=1').data
c.get('/default/json?x=1').data



import os
from warnings import warn
import subprocess
import numpy as np

from ase.calculators.calculator import (Calculator,
                                        FileIOCalculator,
                                        all_changes,
                                        PropertyNotImplementedError)
from ase.units import Bohr, Hartree
from ase.io.xyz import write_xyz
from ase.io.vasp import write_vasp
from ase.parallel import world


class DFTD3(FileIOCalculator):
    """Grimme DFT-D3 calculator"""

    name = 'DFTD3'
    command = 'dftd3'
    dftd3_implemented_properties = ['energy', 'forces', 'stress']

    damping_methods = ['zero', 'bj', 'zerom', 'bjm']

    default_parameters = {'xc': None,  # PBE if no custom damping parameters
                          'grad': True,  # calculate forces/stress
                          'abc': False,  # ATM 3-body contribution
                          'cutoff': 95 * Bohr,  # Cutoff for 2-body calcs
                          'cnthr': 40 * Bohr,  # Cutoff for 3-body and CN calcs
                          'old': False,  # use old DFT-D2 method instead
                          'damping': 'zero',  # Default to zero-damping
                          'tz': False,  # 'triple zeta' alt. parameters
                          's6': None,  # damping parameters start here
                          'sr6': None,
                          's8': None,
                          'sr8': None,
                          'alpha6': None,
                          'a1': None,
                          'a2': None,
                          'beta': None}

    dftd3_flags = ('grad', 'pbc', 'abc', 'old', 'tz')

    def __init__(self,
                 label='ase_dftd3',  # Label for dftd3 output files
                 command=None,  # Command for running dftd3
                 dft=None,  # DFT calculator
                 atoms=None,
                 comm=world,
                 **kwargs):

        self.dft = None
        FileIOCalculator.__init__(self, restart=None,
                                  ignore_bad_restart_file=False,
                                  label=label,
                                  atoms=atoms,
                                  command=command,
                                  dft=dft,
                                  **kwargs)

        self.comm = comm

    def set(self, **kwargs):
        changed_parameters = {}
        # Convert from 'func' keyword to 'xc'. Internally, we only store
        # 'xc', but 'func' is also allowed since it is consistent with the
        # CLI dftd3 interface.
        if kwargs.get('func'):
            if kwargs.get('xc') and kwargs['func'] != kwargs['xc']:
                raise RuntimeError('Both "func" and "xc" were provided! '
                                   'Please provide at most one of these '
                                   'two keywords. The preferred keyword '
                                   'is "xc"; "func" is allowed for '
                                   'consistency with the CLI dftd3 '
                                   'interface.')
            if kwargs['func'] != self.parameters['xc']:
                changed_parameters['xc'] = kwargs['func']
            self.parameters['xc'] = kwargs['func']

        # dftd3 only implements energy, forces, and stresses (for periodic
        # systems). But, if a DFT calculator is attached, and that calculator
        # implements more properties, we will expose those properties too.
        if 'dft' in kwargs:
            dft = kwargs.pop('dft')
            if dft is not self.dft:
                changed_parameters['dft'] = dft
            if dft is None:
                self.implemented_properties = self.dftd3_implemented_properties
            else:
                self.implemented_properties = dft.implemented_properties
            self.dft = dft

        # If the user did not supply an XC functional, but did attach a
        # DFT calculator that has XC set, then we will use that. Note that
        # DFTD3's spelling convention is different from most, so in general
        # you will have to explicitly set XC for both the DFT calculator and
        # for DFTD3 (and DFTD3's will likely be spelled differently...)
        if self.parameters['xc'] is None and self.dft is not None:
            if self.dft.parameters.get('xc'):
                self.parameters['xc'] = self.dft.parameters['xc']

        # Check for unknown arguments. Don't raise an error, just let the
        # user know that we don't understand what they're asking for.
        unknown_kwargs = set(kwargs) - set(self.default_parameters)
        if unknown_kwargs:
            warn('WARNING: Ignoring the following unknown keywords: {}'
                 ''.format(', '.join(unknown_kwargs)))

        changed_parameters.update(FileIOCalculator.set(self, **kwargs))

        # Ensure damping method is valid (zero, bj, zerom, bjm).
        if self.parameters['damping'] is not None:
            self.parameters['damping'] = self.parameters['damping'].lower()
        if self.parameters['damping'] not in self.damping_methods:
            raise ValueError('Unknown damping method {}!'
                             ''.format(self.parameters['damping']))

        # d2 only is valid with 'zero' damping
        elif self.parameters['old'] and self.parameters['damping'] != 'zero':
            raise ValueError('Only zero-damping can be used with the D2 '
                             'dispersion correction method!')

        # If cnthr (cutoff for three-body and CN calculations) is greater
        # than cutoff (cutoff for two-body calculations), then set the former
        # equal to the latter, since that doesn't make any sense.
        if self.parameters['cnthr'] > self.parameters['cutoff']:
            warn('WARNING: CN cutoff value of {cnthr} is larger than '
                 'regular cutoff value of {cutoff}! Reducing CN cutoff '
                 'to {cutoff}.'
                 ''.format(cnthr=self.parameters['cnthr'],
                           cutoff=self.parameters['cutoff']))
            self.parameters['cnthr'] = self.parameters['cutoff']

        # If you only care about the energy, gradient calculations (forces,
        # stresses) can be bypassed. This will greatly speed up calculations
        # in dense 3D-periodic systems with three-body corrections. But, we
        # can no longer say that we implement forces and stresses.
        if not self.parameters['grad']:
            for val in ['forces', 'stress']:
                if val in self.implemented_properties:
                    self.implemented_properties.remove(val)

        # Check to see if we're using custom damping parameters.
        zero_damppars = {'s6', 'sr6', 's8', 'sr8', 'alpha6'}
        bj_damppars = {'s6', 'a1', 's8', 'a2', 'alpha6'}
        zerom_damppars = {'s6', 'sr6', 's8', 'beta', 'alpha6'}
        all_damppars = zero_damppars | bj_damppars | zerom_damppars

        self.custom_damp = False
        damping = self.parameters['damping']
        damppars = set(kwargs) & all_damppars
        if damppars:
            self.custom_damp = True
            if damping == 'zero':
                valid_damppars = zero_damppars
            elif damping in ['bj', 'bjm']:
                valid_damppars = bj_damppars
            elif damping == 'zerom':
                valid_damppars = zerom_damppars

            # If some but not all damping parameters are provided for the
            # selected damping method, raise an error. We don't have "default"
            # values for damping parameters, since those are stored in the
            # dftd3 executable & depend on XC functional.
            missing_damppars = valid_damppars - damppars
            if missing_damppars and missing_damppars != valid_damppars:
                raise ValueError('An incomplete set of custom damping '
                                 'parameters for the {} damping method was '
                                 'provided! Expected: {}; got: {}'
                                 ''.format(damping,
                                           ', '.join(valid_damppars),
                                           ', '.join(damppars)))

            # If a user provides damping parameters that are not used in the
            # selected damping method, let them know that we're ignoring them.
            # If the user accidentally provided the *wrong* set of parameters,
            # (e.g., the BJ parameters when they are using zero damping), then
            # the previous check will raise an error, so we don't need to
            # worry about that here.
            if damppars - valid_damppars:
                warn('WARNING: The following damping parameters are not '
                     'valid for the {} damping method and will be ignored: {}'
                     ''.format(damping,
                               ', '.join(damppars)))

        # The default XC functional is PBE, but this is only set if the user
        # did not provide their own value for xc or any custom damping
        # parameters.
        if self.parameters['xc'] and self.custom_damp:
            warn('WARNING: Custom damping parameters will be used '
                 'instead of those parameterized for {}!'
                 ''.format(self.parameters['xc']))

        if changed_parameters:
            self.results.clear()
        return changed_parameters

    def calculate(self, atoms=None, properties=['energy'],
                  system_changes=all_changes):
        # We don't call FileIOCalculator.calculate here, because that method
        # calls subprocess.call(..., shell=True), which we don't want to do.
        # So, we reproduce some content from that method here.
        Calculator.calculate(self, atoms, properties, system_changes)

        # If a parameter file exists in the working directory, delete it
        # first. If we need that file, we'll recreate it later.
        localparfile = os.path.join(self.directory, '.dftd3par.local')
        if world.rank == 0 and os.path.isfile(localparfile):
            os.remove(localparfile)

        # Write XYZ or POSCAR file and .dftd3par.local file if we are using
        # custom damping parameters.
        self.write_input(self.atoms, properties, system_changes)
        command = self._generate_command()

        # Finally, call dftd3 and parse results.
        # DFTD3 does not run in parallel
        # so we only need it to run on 1 core
        errorcode = 0
        if self.comm.rank == 0:
            with open(self.label + '.out', 'w') as f:
                errorcode = subprocess.call(command,
                                            cwd=self.directory, stdout=f)

        errorcode = self.comm.sum(errorcode)

        if errorcode:
            raise RuntimeError('%s returned an error: %d' %
                               (self.name, errorcode))

        self.read_results()

    def write_input(self, atoms, properties=None, system_changes=None):
        FileIOCalculator.write_input(self, atoms, properties=properties,
                                     system_changes=system_changes)
        # dftd3 can either do fully 3D periodic or non-periodic calculations.
        # It cannot do calculations that are only periodic in 1 or 2
        # dimensions. If the atoms object is periodic in only 1 or 2
        # dimensions, then treat it as a fully 3D periodic system, but warn
        # the user.
        pbc = False
        if any(atoms.pbc):
            if not all(atoms.pbc):
                warn('WARNING! dftd3 can only calculate the dispersion energy '
                     'of non-periodic or 3D-periodic systems. We will treat '
                     'this system as 3D-periodic!')
            pbc = True

        if self.comm.rank == 0:
            if pbc:
                fname = os.path.join(self.directory,
                                     '{}.POSCAR'.format(self.label))
                write_vasp(fname, atoms)
            else:
                fname = os.path.join(
                    self.directory, '{}.xyz'.format(self.label))
                write_xyz(fname, atoms, plain=True)

        # Generate custom damping parameters file. This is kind of ugly, but
        # I don't know of a better way of doing this.
        if self.custom_damp:
            damppars = []
            # s6 is always first
            damppars.append(str(float(self.parameters['s6'])))
            # sr6 is the second value for zero{,m} damping, a1 for bj{,m}
            if self.parameters['damping'] in ['zero', 'zerom']:
                damppars.append(str(float(self.parameters['sr6'])))
            elif self.parameters['damping'] in ['bj', 'bjm']:
                damppars.append(str(float(self.parameters['a1'])))
            # s8 is always third
            damppars.append(str(float(self.parameters['s8'])))
            # sr8 is fourth for zero, a2 for bj{,m}, beta for zerom
            if self.parameters['damping'] == 'zero':
                damppars.append(str(float(self.parameters['sr8'])))
            elif self.parameters['damping'] in ['bj', 'bjm']:
                damppars.append(str(float(self.parameters['a2'])))
            elif self.parameters['damping'] == 'zerom':
                damppars.append(str(float(self.parameters['beta'])))
            # alpha6 is always fifth
            damppars.append(str(int(self.parameters['alpha6'])))
            # last is the version number
            if self.parameters['old']:
                damppars.append('2')
            elif self.parameters['damping'] == 'zero':
                damppars.append('3')
            elif self.parameters['damping'] == 'bj':
                damppars.append('4')
            elif self.parameters['damping'] == 'zerom':
                damppars.append('5')
            elif self.parameters['damping'] == 'bjm':
                damppars.append('6')

            damp_fname = os.path.join(self.directory, '.dftd3par.local')
            if self.comm.rank == 0:
                with open(damp_fname, 'w') as f:
                    f.write(' '.join(damppars))

    def read_results(self):
        # parse the energy
        outname = os.path.join(self.directory, self.label + '.out')
        energy = 0.0
        if self.comm.rank == 0:
            with open(outname, 'r') as f:
                for line in f:
                    if line.startswith(' program stopped'):
                        if 'functional name unknown' in line:
                            message = 'Unknown DFTD3 functional name "{}". ' \
                                      'Please check the dftd3.f source file ' \
                                      'for the list of known functionals ' \
                                      'and their spelling.' \
                                      ''.format(self.parameters['xc'])
                        else:
                            message = 'dftd3 failed! Please check the {} ' \
                                      'output file and report any errors ' \
                                      'to the ASE developers.' \
                                      ''.format(outname)
                        raise RuntimeError(message)

                    if line.startswith(' Edisp'):
                        # line looks something like this:
                        #
                        #     Edisp /kcal,au,ev: xxx xxx xxx
                        #
                        parts = line.split()
                        assert parts[1][0] == '/'
                        index = 2 + parts[1][1:-1].split(',').index('au')
                        e_dftd3 = float(parts[index]) * Hartree
                        energy = e_dftd3
                        break
                else:
                    raise RuntimeError('Could not parse energy from dftd3 '
                                       'output, see file {}'.format(outname))

        self.results['energy'] = self.comm.sum(energy)
        self.results['free_energy'] = self.results['energy']

        # FIXME: Calculator.get_potential_energy() simply inspects
        # self.results for the free energy rather than calling
        # Calculator.get_property('free_energy'). For example, GPAW does
        # not actually present free_energy as an implemented property, even
        # though it does calculate it. So, we are going to add in the DFT
        # free energy to our own results if it is present in the attached
        # calculator. TODO: Fix the Calculator interface!!!
        if self.dft is not None:
            try:
                efree = self.dft.get_potential_energy(
                    force_consistent=True)
                self.results['free_energy'] += efree
            except PropertyNotImplementedError:
                pass

        if self.parameters['grad']:
            # parse the forces
            forces = np.zeros((len(self.atoms), 3))
            forcename = os.path.join(self.directory, 'dftd3_gradient')
            if self.comm.rank == 0:
                with open(forcename, 'r') as f:
                    for i, line in enumerate(f):
                        forces[i] = np.array([float(x) for x in line.split()])
                forces *= -Hartree / Bohr
            self.comm.broadcast(forces, 0)
            self.results['forces'] = forces

            if any(self.atoms.pbc):
                # parse the stress tensor
                stress = np.zeros((3, 3))
                stressname = os.path.join(self.directory, 'dftd3_cellgradient')
                if self.comm.rank == 0:
                    with open(stressname, 'r') as f:
                        for i, line in enumerate(f):
                            for j, x in enumerate(line.split()):
                                stress[i, j] = float(x)
                    stress *= Hartree / Bohr / self.atoms.get_volume()
                    stress = np.dot(stress, self.atoms.cell.T)
                self.comm.broadcast(stress, 0)
                self.results['stress'] = stress.flat[[0, 4, 8, 5, 2, 1]]

    def get_property(self, name, atoms=None, allow_calculation=True):
        dft_result = None
        if self.dft is not None:
            dft_result = self.dft.get_property(name, atoms, allow_calculation)

        dftd3_result = FileIOCalculator.get_property(self, name, atoms,
                                                     allow_calculation)

        if dft_result is None and dftd3_result is None:
            return None
        elif dft_result is None:
            return dftd3_result
        elif dftd3_result is None:
            return dft_result
        else:
            return dft_result + dftd3_result

    def _generate_command(self):
        command = self.command.split()

        if any(self.atoms.pbc):
            command.append(self.label + '.POSCAR')
        else:
            command.append(self.label + '.xyz')

        if not self.custom_damp:
            xc = self.parameters.get('xc')
            if xc is None:
                xc = 'pbe'
            command += ['-func', xc.lower()]

        for arg in self.dftd3_flags:
            if self.parameters.get(arg):
                command.append('-' + arg)

        if any(self.atoms.pbc):
            command.append('-pbc')

        command += ['-cnthr', str(self.parameters['cnthr'] / Bohr)]
        command += ['-cutoff', str(self.parameters['cutoff'] / Bohr)]

        if not self.parameters['old']:
            command.append('-' + self.parameters['damping'])

        return command



import numpy as np
from ase.build import bulk, molecule
from ase.io.jsonio import encode, decode


def assert_equal(atoms1, atoms2):
    assert atoms1 == atoms2
    assert set(atoms1.arrays) == set(atoms2.arrays)
    for name in atoms1.arrays:
        assert np.array_equal(atoms1.arrays[name], atoms2.arrays[name]), name


atoms = bulk('Ti')
print('atoms', atoms)
txt = encode(atoms)
print('encoded', txt)

atoms1 = decode(txt)
print('decoded', atoms1)
txt1 = encode(atoms1)
assert txt == txt1
assert_equal(atoms, atoms1)


BeH = molecule('BeH')
assert BeH.has('initial_magmoms')
new_BeH = decode(encode(BeH))
assert_equal(BeH, new_BeH)
assert new_BeH.has('initial_magmoms')

from ase.constraints import FixAtoms
atoms = bulk('Ti')
atoms.constraints = FixAtoms(indices=[0])
newatoms = decode(encode(atoms))
c1 = atoms.constraints
c2 = newatoms.constraints
assert len(c1) == len(c2) == 1
# Can we check constraint equality somehow?
# Would make sense for FixAtoms
assert np.array_equal(c1[0].index, c2[0].index)
